"""
Instrument driver for a Newport single axis motion controller. This is a base class for other controllers,
but can be used without extending.
"""
from contextlib import contextmanager
import enum
import logging
from time import sleep
from typing import Dict, List, Optional, Tuple, Union

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport
from qmi.instruments.newport.actuators import LinearActuator


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class HomeSearchTypes(enum.IntEnum):
    """The types of HOME search used with the OR command."""
    MZ_SWITCH_AND_ENCODER_INDEX = 0
    CURRENT_POSITION_AS_HOME = 1
    MZ_SWITCH_ONLY = 2
    EOR_SWITCH_AND_ENCODER_INDEX = 3
    EOR_SWITCH_ONLY = 4


class Newport_SingleAxisMotionController(QMI_Instrument):
    """
    Instrument driver for a Newport single-axis motion controller. This device
    is the controller for an actuator and is controlled via serial.
    """

    # Default response timeout in seconds.
    DEFAULT_RESPONSE_TIMEOUT = 5.0

    # Default home search time out
    DEFAULT_HOME_SEARCH_TIMEOUT = 5.0

    # Time to execute a command
    COMMAND_EXEC_TIME = 0.5
    PW0_EXEC_TIME = 10

    # Code for no error
    ERROR_CODE_NONE = "@"

    # Default contoller address to send commands to.
    DEFAULT_CONTROLLER_ADDRESS = 1

    # Floating point value FW limits
    MIN_FLOAT_LIMIT = 1E-6
    MAX_FLOAT_LIMIT = 1E12

    STATE_TABLE = {
        "0A": "NOT REFERENCED from RESET",
        "0B": "NOT REFERENCED from HOMING.",
        "0C": "NOT REFERENCED from CONFIGURATION.",
        "0D": "NOT REFERENCED from DISABLE.",
        "0E": "NOT REFERENCED from READY.",
        "0F": "NOT REFERENCED from MOVING.",
        "10": "NOT REFERENCED - NO PARAMETERS IN MEMORY.",
        "14": "CONFIGURATION.",
        "1E": "HOMING.",
        "28": "MOVING.",
        "32": "READY from HOMING.",
        "33": "READY from MOVING.",
        "34": "READY from DISABLE.",
        "36": "READY T from READY.",
        "37": "READY T from TRACKING.",
        "38": "READY T from DISABLE T.",
        "3C": "DISABLE from READY.",
        "3D": "DISABLE from MOVING.",
        "3E": "DISABLE from TRACKING.",
        "3F": "DISABLE from READY T.",
        "46": "TRACKING from READY T.",
        "47": "TRACKING from TRACKING.",
    }

    POSITIONER_ERROR_TABLE = {
        6: "80W output power exceeded",
        7: "DC voltage too low",
        8: "Wrong ESP stage",
        9: "Homing time out",
        10: "Following error",
        11: "Short circuit detection",
        12: "RMS current limit",
        13: "Peak current limit",
        14: "Positive end of run",
        15: "Negative end of run",
    }

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str,
                 serial: str,
                 actuators: Dict[Optional[int], LinearActuator],
                 baudrate: int) -> None:
        """Initialize driver.

        Parameters:
            name:                   Name for this instrument instance.
            transport:              QMI transport descriptor to connect to the instrument.
            serial:                 The serial number of the instrument.
            actuators:              The linear actuators that this controller will drive. Each controller address
                                    drives a linear actuator. The key of the dictionary is the controller address
                                    and the value is the actuator that it drives.
            baudrate:               The baudrate of the instrument.
        """
        super().__init__(context, name)
        self._timeout = self.DEFAULT_RESPONSE_TIMEOUT
        self._transport_str = transport
        self._transport = create_transport(
            transport, default_attributes={"baudrate": baudrate})
        self._scpi_protocol = ScpiProtocol(self._transport,
                                           command_terminator="\r\n",
                                           response_terminator="\r\n",
                                           default_timeout=self._timeout)
        self._serial = serial
        self._actuators = actuators
        self._controller_address: int = self.DEFAULT_CONTROLLER_ADDRESS

    @property
    def controller_address(self) -> Optional[int]:
        """Address of the controller that needs to be controlled."""
        return self._controller_address

    @controller_address.setter
    def controller_address(self, controller_address: Optional[int]) -> None:
        # if controller address is not given use the default one.
        self._controller_address = controller_address or self.DEFAULT_CONTROLLER_ADDRESS
        # Check that the address is valid.
        if self._controller_address not in self._actuators.keys():
            raise QMI_InstrumentException(
                f"Invalid controller address {self._controller_address}")

    @contextmanager
    def configuration_state(self):
        try:
            yield self._enter_configuration_state()
        finally:
            self._exit_configuration_state()

    def _build_command(
            self, cmd: str, value: Union[float, int, None] = None
    ) -> str:
        """Build the command.

        Parameters:
            cmd:                Name of command.
            value:              Value to go with command if needed.
        """
        controller_address = self.controller_address
        return f"{controller_address}{cmd}{value}\r\n" if value is not None else f"{controller_address}{cmd}\r\n"

    def _check_error(self) -> None:
        """
        Check the currently memorised error.

        Raises:
            QMI_InstrumentException if an error exists.
        """
        error_code = self._scpi_protocol.ask(
            self._build_command("TE"))[-1]
        sleep(self.COMMAND_EXEC_TIME)
        if error_code != self.ERROR_CODE_NONE:
            error_str = self._scpi_protocol.ask(self._build_command(
                "TB" + error_code)).strip().split(' ', 1)[1]
            raise QMI_InstrumentException(f"Error {error_code}: {error_str}")

    def _get_current_state(self) -> str:
        """
        Get current controller state.

        Returns:
            state: The current controller 16-bit state value string.
        """
        err_and_state = self._scpi_protocol.ask(
            self._build_command("TS")
        )
        return err_and_state[7:9]

    def _state_ready_check(self, parameter: str) -> int:
        """Check if the state is READY (32-38) or DISABLE (3C-3F). Otherwise, raise an exception.

        Parameters:
            parameter: String to describe for which parameter the check is made.

        Raises:
            QMI_InstrumentException: If the state is not READY or DISABLE, with informing the parameter name.

        Returns:
            istate: The state number as an integer.
        """
        state = self._get_current_state()
        istate = int(state, 16)
        if istate < int("32", 16) or istate > int("3F", 16):
            raise QMI_InstrumentException(
                f"Cannot set {parameter} in controller state {self.STATE_TABLE[state]}")

        return istate

    def _enter_configuration_state(self) -> None:
        """
        Enter the CONFIGURATION state of the controller from the NOT REFERENCED state. If the controller
        is not in the NOT REFERENCED state, then call the `reset` method.
        NOTE:   In this state the parameters are stored in the flash memory of the controller.
                The device supports up to 100 writes, so this command should not be used often.
        """
        state = self._get_current_state()
        istate = int(state, 16)
        if istate == int("14", 16):
            # Already in CONFIGURATION state
            return

        _logger.info("Entering configuration state of instrument [%s]", self._name)
        # If the state is not NOT REFERENCED (0A-10) or CONFIGURATION (14), then reset.
        if istate > int("14", 16):
            # Call reset to get to NOT REFERENCED state.
            self._scpi_protocol.write(
                self._build_command("RS")
            )
            sleep(self.COMMAND_EXEC_TIME)

        self._scpi_protocol.write(
            self._build_command("PW", 1)
        )
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()

    def _exit_configuration_state(self) -> None:
        """
        Exit the CONFIGURATION state of the controller to the NOT REFERENCED state.
        NOTE:   The execution of a PW0 command may take up to 10 seconds. During that time the
                controller will not respond to any other command.
        """
        _logger.info("Exiting configuration state of instrument [%s]", self._name)
        self._scpi_protocol.write(
            self._build_command("PW", 0)
        )
        sleep(self.PW0_EXEC_TIME)

    def _enter_disable_state(self) -> None:
        """
        Enter the DISABLE state of the controller from the READY state. If the controller
        is not in the NOT REFERENCE state, then call the `reset` method.
        NOTE:   In this state the parameters are stored in the flash memory of the controller.
                The device supports up to 100 writes, so this command should not be used often.
        """
        _logger.info("Entering disable state of instrument [%s]", self._name)
        self._scpi_protocol.write(
            self._build_command("MM", 1)
        )
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()

    def _exit_disable_state(self) -> None:
        """
        Exit the DISABLE state of the controller to the READY state.
        """
        _logger.info("Exiting disable state of instrument [%s]", self._name)
        self._scpi_protocol.write(
            self._build_command("MM", 0)
        )

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection to instrument [%s]", self._name)
        self._check_is_closed()
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to instrument [%s]", self._name)
        self._check_is_open()
        self._transport.close()
        super().close()

    @rpc_method
    def reset(self, controller_address: Optional[int] = None) -> None:
        """
        Reset the instrument. Equivalent to a power-up.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        _logger.info("Resetting instrument [%s]", self._name)
        self._scpi_protocol.write(
            self._build_command("RS"))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()

    @rpc_method
    def get_idn(self, controller_address: Optional[int] = None) -> QMI_InstrumentIdentification:
        """
        Read instrument type and version and return QMI_InstrumentIdentification instance.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            QMI_InstrumentIdentification with the information of the instrument.
        """
        _logger.info("Getting identification of instrument [%s]", self._name)
        self.controller_address = controller_address
        instr_info = self._scpi_protocol.ask(
            self._build_command("VE"))
        self._check_error()
        words = instr_info[1:].strip().split()
        return QMI_InstrumentIdentification(vendor="Newport",
                                            model=words[1],
                                            version=words[2],
                                            serial=self._serial)

    @rpc_method
    def get_stage_identifier(self, controller_address: Optional[int] = None) -> str:
        """
        Read stage identifier.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            The stage identifier.
        """
        _logger.info("Getting stage identifier of instrument [%s]", self._name)
        self.controller_address = controller_address
        # instrument must be in CONFIGURATION state to get the stage identifier.
        with self.configuration_state():
            identifier = self._scpi_protocol.ask(
                self._build_command("ID?"))
            self._check_error()

        return identifier

    @rpc_method
    def get_positioner_error_and_state(self, controller_address: Optional[int] = None) -> Tuple[List[str], str]:
        """
        Get the positioner error and the current state of the controller.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        _logger.info(
            "Getting positioner error and state of instrument [%s]", self._name)
        self.controller_address = controller_address
        err_and_state = self._scpi_protocol.ask(
            self._build_command("TS"))
        err = err_and_state[3:7]
        state = err_and_state[7:9]
        errors = []
        # convert the hexadecimal error code to binary
        bin_err_str = "{0:016b}".format(int(err, 16))
        # loop over bits and log the errors (i.e. high bits)
        for i, bit in enumerate(bin_err_str):
            if bit == "1":
                errors.append(self.POSITIONER_ERROR_TABLE[i])

        return errors, self.STATE_TABLE[state]

    @rpc_method
    def home_search(self, controller_address: Optional[int] = None) -> None:
        """
        Execute the home search. This is needed before any motion commands can be executed. It finds an origin position
        for the actuator. Can be done only in NOT REFERENCED state.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        _logger.info("Homing instrument [%s]", self._name)
        self.controller_address = controller_address
        self._scpi_protocol.write(
            self._build_command("OR"))
        self._check_error()

    @rpc_method
    def set_home_search_timeout(
            self, timeout: Optional[float] = None, controller_address: Optional[int] = None) -> None:
        """
        Set the timeout for the home search.

        Parameters:
            timeout:            Optional timeout in seconds. If not set, it defaults DEFAULT_HOME_SEARCH_TIMEOUT
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        t = timeout if timeout else self.DEFAULT_HOME_SEARCH_TIMEOUT
        _logger.info(
            "Setting homing timeout of instrument [%s] to [%s]", self._name, t)
        self.controller_address = controller_address
        # instrument must be in CONFIGURATION state to set the timeout.
        with self.configuration_state():
            self._scpi_protocol.write(
                self._build_command("OT", t))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

    @rpc_method
    def get_home_search_timeout(self, controller_address: Optional[int] = None) -> float:
        """
        Get the timeout for the home search.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            The timeout in seconds.
        """
        _logger.info("Getting homing timeout of instrument [%s]", self._name)
        self.controller_address = controller_address
        timeout = self._scpi_protocol.ask(
            self._build_command("OT?"))
        self._check_error()
        return float(timeout[3:])

    @rpc_method
    def move_absolute(self, position: float, controller_address: Optional[int] = None) -> None:
        """
        Perform an absolute move. This command can take several seconds to finish. However, it is not blocking.
        Use other methods such as `get_positioner_error_and_state` to query the state of the controller.

        Parameters:
            position:           New position to move to, in encoder units. Must be within actuator's travel range.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        if position > self._actuators[self.controller_address].TRAVEL_RANGE.max or \
                position < self._actuators[self.controller_address].TRAVEL_RANGE.min:
            raise QMI_InstrumentException(
                f"Provided value {position} outside valid range ["
                f"{self._actuators[self.controller_address].TRAVEL_RANGE.min}, "
                f"{self._actuators[self.controller_address].TRAVEL_RANGE.max}]"
            )

        # The move can be made only in READY state.
        state = self._state_ready_check("absolute move")
        if state > int("38", 16):
            raise QMI_InstrumentException("The controller must be in READY state to move.")

        _logger.info(
            "Performing an absolute move of instrument [%s] to [%s]", self._name, position)
        self._scpi_protocol.write(
            self._build_command("PA", position))
        self._check_error()

    @rpc_method
    def get_setpoint(self, controller_address: Optional[int] = None) -> float:
        """
        Get the set-point position of the actuator according to the encoder value.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            Set-point position in encoder units.
        """
        _logger.info("Getting set-point of instrument [%s]", self._name)
        self.controller_address = controller_address
        setpoint = self._scpi_protocol.ask(
            self._build_command("TH"))
        self._check_error()
        return float(setpoint[3:])

    @rpc_method
    def get_position(self, controller_address: Optional[int] = None) -> float:
        """
        Get the actual position of the actuator according to the encoder value.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            Current position in encoder units.
        """
        _logger.info("Getting position of instrument [%s]", self._name)
        self.controller_address = controller_address
        pos = self._scpi_protocol.ask(
            self._build_command("TP"))
        self._check_error()
        return float(pos[3:])

    @rpc_method
    def move_relative(self, displacement: float, controller_address: Optional[int] = None) -> None:
        """
        Perform a relative move from the current position.

        Parameters:
            displacement:       Displacement from current position.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        if abs(displacement) < self._actuators[self.controller_address].MIN_INCREMENTAL_MOTION:
            raise QMI_InstrumentException(
                f"Provided value {displacement} lower than minimum "
                f"{self._actuators[self.controller_address].MIN_INCREMENTAL_MOTION}"
            )

        # The move can be made only in READY state.
        state = self._state_ready_check("relative move")
        if state > int("38", 16):
            raise QMI_InstrumentException("The controller must be in READY state to move.")

        _logger.info("Performing relative move of instrument [%s]", self._name)
        self._scpi_protocol.write(
            self._build_command("PR", displacement))
        self._check_error()

    @rpc_method
    def get_motion_time(self, displacement: float, controller_address: Optional[int] = None) -> float:
        """
        Get the motion time for a relative move.

        Parameters:
            displacement:       Displacement from current position (relative move size).
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            Motion time for a relative move in seconds.
        """
        _logger.info("Getting motion time for a relative move of instrument [%s]", self._name)
        self.controller_address = controller_address
        move_time = self._scpi_protocol.ask(
            self._build_command("PT", displacement))
        self._check_error()
        return float(move_time[3:])

    @rpc_method
    def is_in_configuration_state(self, controller_address: Optional[int] = None) -> bool:
        """
        Get the CONFIGURATION state of the controller.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            state: Boolean indicating if the controller is in CONFIGURATION state (True) or not (False).
        """
        _logger.info("Getting CONFIGURATION state of instrument [%s]", self._name)
        self.controller_address = controller_address
        state = self._scpi_protocol.ask(
            self._build_command("PW?"))
        self._check_error()
        return bool(int(state[3:]))

    @rpc_method
    def set_configuration_state(self, state: bool, controller_address: Optional[int] = None) -> None:
        """
        Set the NOT REFERENCED or CONFIGURATION state of the controller.
        If the controller is not in the either state, then an exception is raised in the _check_error.
        NOTE: In this state the parameters are stored in the flash memory of the controller.

        Parameters:
            state:              False for setting the configuration state from CONFIGURATION to NOT REFERENECED or
                                True for setting the configuration state from NOT REFERENECED to CONFIGURATION.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        """
        _logger.info(
            "Setting state of instrument [%s] to [%s]", self._name, "CONFIGURATION" if state else "NOT REFERENCED"
        )
        self.controller_address = controller_address
        self._scpi_protocol.write(
            self._build_command("PW", int(state)))

        if not state:
            # The execution of a PW0 command may take up to 10 seconds. During that time the
            # controller will not respond to any other command.
            sleep(self.PW0_EXEC_TIME)

        else:
            sleep(self.COMMAND_EXEC_TIME)

    @rpc_method
    def is_in_disable_state(self, controller_address: Optional[int] = None) -> bool:
        """
        Get the DISABLE or READY state of the controller.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            state: Boolean to indicate state is DISABLE (True) or READY (False)
        """
        _logger.info("Getting DISABLE state of instrument [%s]", self._name)
        self.controller_address = controller_address
        state = self._scpi_protocol.ask(
            self._build_command("MM?"))
        self._check_error()
        return bool(int(state[3:]))

    @rpc_method
    def set_disable_state(self, state: bool, controller_address: Optional[int] = None) -> None:
        """
        Set the DISABLE or READY state of the controller from the READY or DISABLE state, respectively.
        If the controller is not in the either state, then an exception is raised in the _check_error.
        NOTE: In this state the parameters are stored in the flash memory of the controller.

        Parameters:
            state: True for setting state from READY to DISABLE, False for vice versa.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        _logger.info("Setting state of instrument [%s] to [%s]", self._name, "DISABLE" if state else "READY")
        self.controller_address = controller_address
        self._scpi_protocol.write(
            self._build_command("MM", int(state)))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()

    @rpc_method
    def get_acceleration(self, controller_address: Optional[int] = None) -> float:
        """
        Get the acceleration of the actuator in preset unit/s^2, so if the encoder unit is mm,
        then the returned value is in mm/s^2.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            Encoder increment value.
        """
        def _get_acceleration() -> str:
            acceleration = self._scpi_protocol.ask(
                self._build_command("AC?"))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()
            return acceleration

        _logger.info("Getting acceleration of instrument [%s]", self._name)
        self.controller_address = controller_address
        # instrument must be in CONFIGURATION, DISABLE or READY state to get the acceleration.
        try:
            self._state_ready_check("acceleration")
            acceleration = _get_acceleration()

        except QMI_InstrumentException:
            with self.configuration_state():
                acceleration = _get_acceleration()

        return float(acceleration[3:])

    @rpc_method
    def set_acceleration(
            self, acceleration: float, persist: bool = False, controller_address: Optional[int] = None
    ) -> None:
        """
        Set the acceleration at which the actuator moves. Can be set only in CONFIGURATION, READY and DISABLE states.

        Parameters:
            acceleration:       Acceleration in preset unit/s^2. The unit depends on the encoder resolution,
                                which is usually set to 1mm.
            persist:            Flag to indicate if the acceleration should be persisted to the controller's memory, so
                                it is still available after powering down the controller. When not persisted, the
                                maximum allowable acceleration that can be set is the one stored in the controller's
                                memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        def _set_acceleration():
            self._scpi_protocol.write(
                self._build_command("AC", acceleration))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

        if self.MIN_FLOAT_LIMIT >= acceleration or acceleration >= self.MAX_FLOAT_LIMIT:
            raise QMI_InstrumentException(
                f"Provided value {acceleration} not in valid range {self.MIN_FLOAT_LIMIT} "
                f"< acceleration < {self.MAX_FLOAT_LIMIT}."
            )

        _logger.info(
            "Setting acceleration of instrument [%s] to [%f]", self._name, acceleration)
        self.controller_address = controller_address
        if persist:
            # instrument must be in CONFIGURATION state to set persistent acceleration.
            with self.configuration_state():
                _set_acceleration()

        else:
            # instrument must be in DISABLE or READY state to set acceleration.
            self._state_ready_check("acceleration")
            _set_acceleration()

    @rpc_method
    def get_velocity(self, controller_address: Optional[int] = None) -> float:
        """
        Get the velocity of the actuator in unit/s, so if the encoder unit is mm,
        then the returned value is in mm/s.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        def _get_velocity() -> str:
            velocity = self._scpi_protocol.ask(
                self._build_command("VA?"))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()
            return velocity

        _logger.info("Getting velocity of instrument [%s]", self._name)
        self.controller_address = controller_address
        # instrument must be in CONFIGURATION, DISABLE or READY state to get the velocity.
        try:
            self._state_ready_check("velocity")
            velocity = _get_velocity()

        except QMI_InstrumentException:
            with self.configuration_state():
                velocity = _get_velocity()

        return float(velocity[3:])

    @rpc_method
    def set_velocity(self, velocity: float, persist: bool = False, controller_address: Optional[int] = None) -> None:
        """
        Set the velocity at which the actuator moves.

        Parameters:
            velocity:           Velocity in unit/s. The unit depends on the encoder resolution,
                                which is usually set to 1mm.
            persist:            Flag to indicate if the velocity should be persisted to the controller's memory, so it
                                is still available after powering down the controller. When not persisted, the maximum
                                allowable velocity that can be set is the one stored in the controller's memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        def _set_velocity():
            self._scpi_protocol.write(
                self._build_command("VA", velocity))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

        self.controller_address = controller_address
        if velocity > self._actuators[self.controller_address].MAX_VELOCITY:
            raise QMI_InstrumentException(
                f"Provided value {velocity} greater than allowed maximum "
                f"{self._actuators[self.controller_address].MAX_VELOCITY}"
            )
        if velocity < self._actuators[self.controller_address].MIN_VELOCITY:
            raise QMI_InstrumentException(
                f"Provided value {velocity} lower than minimum "
                f"{self._actuators[self.controller_address].MIN_VELOCITY}"
            )

        _logger.info(
            "Setting velocity of instrument [%s] to [%f]", self._name, velocity)
        if persist:
            # instrument must be in CONFIGURATION state to set persistent velocity.
            with self.configuration_state():
                _set_velocity()

        else:
            # instrument must be in DISABLE or READY state to set velocity.
            self.controller_address = controller_address
            self._state_ready_check("velocity")
            _set_velocity()

    @rpc_method
    def get_jerk_time(self, controller_address: Optional[int] = None) -> float:
        """
        Get the jerk time of the actuator.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            jerk_time: Jerk time in seconds.
        """
        def _get_jerk_time() -> str:
            jerk_time = self._scpi_protocol.ask(
                self._build_command("JR?"))
            self._check_error()
            return jerk_time

        _logger.info("Getting jerk time of instrument [%s]", self._name)
        self.controller_address = controller_address
        # instrument must be in CONFIGURATION, DISABLE or READY state to get the jerk time.
        try:
            self._state_ready_check("jerk time")
            jerk_time = _get_jerk_time()

        except QMI_InstrumentException:
            with self.configuration_state():
                jerk_time = _get_jerk_time()

        return float(jerk_time[3:])

    @rpc_method
    def set_jerk_time(self, jerk_time: float, persist: bool = False, controller_address: Optional[int] = None) -> None:
        """
        Set the jerk time at which the actuator accelerates.

        Parameters:
            jerk_time:          Jerk time in seconds.
            persist:            Flag to indicate if the jerk time should be persisted to the controller's memory, so it
                                is still available after powering down the controller. When not persisted, the maximum
                                allowable jerk time that can be set is the one stored in the controller's memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        def _set_jerk_time():
            self._scpi_protocol.write(self._build_command(
                "JR", jerk_time))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

        if jerk_time <= 0.001 or jerk_time >= self.MAX_FLOAT_LIMIT:
            raise QMI_InstrumentException(
                f"Provided value {jerk_time} not in valid range 0.001 < jerk_time < {self.MAX_FLOAT_LIMIT}")

        _logger.info(
            "Setting jerk time of instrument [%s] to [%f]", self._name, jerk_time)
        self.controller_address = controller_address
        if persist:
            # instrument must be in CONFIGURATION state to set persistent jerk time.
            with self.configuration_state():
                _set_jerk_time()

        else:
            # instrument must be in DISABLE or READY state to set jerk time.
            self._state_ready_check("cut-off frequency")
            _set_jerk_time()

    @rpc_method
    def get_error(self, controller_address: Optional[int] = None) -> Tuple[str, str]:
        """
        Get the currently memorised error.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            A tuple containing the error code and the human-readable error message.
        """
        _logger.info("Getting error of instrument [%s]", self._name)
        self.controller_address = controller_address
        error_code = self._scpi_protocol.ask(
            self._build_command("TE"))[-1]
        if error_code == self.ERROR_CODE_NONE:
            return "@", "No error"

        error_str = self._scpi_protocol.ask(self._build_command(
            "TB" + error_code)).strip().split(' ', 1)[1]
        return error_code, error_str

    @rpc_method
    def stop_motion(self, controller_address: Optional[int] = None) -> None:
        """
        Stop the motion of the actuator by decelerating it. Works on DISABLE, READY and MOTION states.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        _logger.info("Stop motion of instrument [%s]", self._name)
        self.controller_address = controller_address
        self._scpi_protocol.write(
            self._build_command("ST"))
        self._check_error()

    @rpc_method
    def get_backlash_compensation(self, controller_address: Optional[int] = None) -> float:
        """
        Get the backlash value in encoder units.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        _logger.info("Getting backlash compensation of controller [%s] instrument [%s]",
                     self.controller_address, self._name)
        # instrument must be in CONFIGURATION state to get the backlash compensation.
        with self.configuration_state():
            backlash_comp = self._scpi_protocol.ask(
                self._build_command("BA?"))
            self._check_error()

        return float(backlash_comp[3:])

    @rpc_method
    def set_backlash_compensation(self, backlash_comp: float, controller_address: Optional[int] = None) -> None:
        """
        Set the backlash compensation of a controller.

        Parameters:
            backlash_comp:      backlash compensation in encoder units.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        # instrument must be in CONFIGURATION state to set the backlash compensation.
        with self.configuration_state():
            # First check if the hysteresis compensation is enabled.
            hysteresis_comp = self._scpi_protocol.ask(
                self._build_command("BH?"))
            self._check_error()
            if float(hysteresis_comp[3:]) > 0.0:
                self._exit_configuration_state()
                raise QMI_InstrumentException(
                    "Backlash compensation cannot be set if hysteresis compensation is enabled!"
                )

            _logger.info(
                "Setting backlash compensation of controller [%s] instrument [%s] to [%f]",
                self.controller_address, self._name, backlash_comp)
            self._scpi_protocol.write(
                self._build_command("BA", backlash_comp))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

    @rpc_method
    def get_hysteresis_compensation(self, controller_address: Optional[int] = None) -> float:
        """
        Get the hysteresis value in encoder units.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        _logger.info("Getting hysteresis compensation of controller [%s] instrument [%s]",
                     self.controller_address, self._name)
        # instrument must be in CONFIGURATION state to get the hysteresis compensation.
        with self.configuration_state():
            hysteresis_comp = self._scpi_protocol.ask(
                self._build_command("BH?"))
            self._check_error()

        return float(hysteresis_comp[3:])

    @rpc_method
    def set_hysteresis_compensation(self, hysteresis_comp: float, controller_address: Optional[int] = None) -> None:
        """
        Set the hysteresis compensation of a controller.

        Parameters:
            hysteresis_comp:      hysteresis compensation in encoder units.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        # instrument must be in CONFIGURATION state to set the hysteresis compensation.
        with self.configuration_state():
            # First check if the backlash compensation is enabled.
            backlash_comp = self._scpi_protocol.ask(
                self._build_command("BA?"))
            self._check_error()
            if float(backlash_comp[3:]) > 0.0:
                self._exit_configuration_state()
                raise QMI_InstrumentException(
                        "Hysteresis compensation cannot be set if backlash compensation is enabled!")

            _logger.info(
                "Setting hysteresis compensation of controller [%s] instrument [%s] to [%f]",
                self.controller_address, self._name, hysteresis_comp)
            self._scpi_protocol.write(
                self._build_command("BH", hysteresis_comp))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

    @rpc_method
    def get_home_search_type(self, controller_address: Optional[int] = None) -> int:
        """
        Get the type of HOME search used with the OR command.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting the type of HOME search used with the OR command of instrument [%s]", self._name)
        # instrument must be in CONFIGURATION state to get the home search type.
        with self.configuration_state():
            home_search_type = self._scpi_protocol.ask(
                self._build_command("HT?"))
            self._check_error()

        return int(home_search_type[3:])

    @rpc_method
    def set_home_search_type(self, home_search_type: int, controller_address: Optional[int] = None) -> None:
        """
        Set the type of HOME search used with the OR command.

        Parameters:
            home_search_type:   New type of HOME search used with the OR command.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        hst_values = [hst.value for hst in set(HomeSearchTypes)]
        if home_search_type not in hst_values:
            raise QMI_InstrumentException(
                f"Provided value {home_search_type} not in valid range {hst_values}.")

        _logger.info(
            "Setting the type of HOME search used with the OR command for instrument [%s] to [%s]",
            self._name, HomeSearchTypes(home_search_type).name
        )
        self.controller_address = controller_address
        # instrument must be in CONFIGURATION state to set the home search type.
        with self.configuration_state():
            self._scpi_protocol.write(self._build_command(
                "HT", home_search_type))
            self._check_error()

    @rpc_method
    def get_peak_current_limit(self, controller_address: Optional[int] = None) -> float:
        """
        Get the controller’s maximum or peak output current limit to the motor.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        _logger.info("Getting the peak current limit of controller [%s] instrument [%s]",
                     self.controller_address, self._name)
        # instrument must be in CONFIGURATION state to get the current limit.
        with self.configuration_state():
            current_limit = self._scpi_protocol.ask(
                self._build_command("QIL?"))
            self._check_error()

        return float(current_limit[4:])

    @rpc_method
    def set_peak_current_limit(self, current_limit: float, controller_address: Optional[int] = None) -> None:
        """
        Set the controller’s maximum or peak output current limit to the motor.

        Parameters:
            current_limit:    Controller’s maximum or peak output current limit to the motor in [A].
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        if current_limit < 0.05 or current_limit > 3.0:
            raise QMI_InstrumentException(
                    "Current limit value not in valid range 0.05 <= current_limit <= 3.0")

        self.controller_address = controller_address
        _logger.info(
            "Setting peak current limit of controller [%s] instrument [%s] to [%f]",
            self.controller_address, self._name, current_limit)
        # instrument must be in CONFIGURATION state to set the current limit.
        with self.configuration_state():
            self._scpi_protocol.write(
                self._build_command("QIL", current_limit))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

    @rpc_method
    def get_rms_current_limit(self, controller_address: Optional[int] = None) -> float:
        """
        Get the controller’s rms output current limit to the motor.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        _logger.info("Getting the RMS current limit of controller [%s] instrument [%s]",
                     self.controller_address, self._name)
        # instrument must be in CONFIGURATION state to get the current limit.
        with self.configuration_state():
            current_limit = self._scpi_protocol.ask(
                self._build_command("QIR?"))
            self._check_error()

        return float(current_limit[4:])

    @rpc_method
    def set_rms_current_limit(self, current_limit: float, controller_address: Optional[int] = None) -> None:
        """
        Set the controller’s maximum or rms output current limit to the motor.

        Parameters:
            current_limit:    Controller’s maximum or rms output current limit to the motor in [A].
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        # instrument must be in CONFIGURATION state to check the peak current limit and to set the RMS current limit.
        with self.configuration_state():
            peak_current_limit = self._scpi_protocol.ask(
                self._build_command("QIL?"))
            self._check_error()
            peak_current_limit_f = min(1.5, float(peak_current_limit[4:]))
            if current_limit < 0.05 or current_limit > peak_current_limit_f:
                self._exit_configuration_state()
                raise QMI_InstrumentException(
                        f"Current limit value not in valid range 0.05 <= current_limit <= {peak_current_limit_f}"
                )

            _logger.info(
                "Setting RMS current limit of controller [%s] instrument [%s] to [%f]",
                self.controller_address, self._name, current_limit)
            self._scpi_protocol.write(self._build_command("QIR", current_limit))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

    @rpc_method
    def get_rms_current_averaging_time(self, controller_address: Optional[int] = None) -> float:
        """
        Get the controller’s averaging period for rms current calculation.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        # instrument must be in CONFIGURATION state to get the current limit.
        with self.configuration_state():
            _logger.info("Getting the averaging period for rms current calculation of controller [%s] instrument [%s]",
                         self.controller_address, self._name)
            averaging_time = self._scpi_protocol.ask(
                self._build_command("QIT?"))
            self._check_error()

        return float(averaging_time[4:])

    @rpc_method
    def set_rms_current_averaging_time(self, averaging_time: float, controller_address: Optional[int] = None) -> None:
        """
        Set the controller’s averaging period for rms current calculation.

        Parameters:
            averaging_time:    Controller’s averaging period for rms current calculation in [s].
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        if averaging_time <= 0.01 or averaging_time > 100.0:
            raise QMI_InstrumentException(
                    "Averaging period for rms current calculation not in valid range 0.01 < averaging_time <= 100.0")

        self.controller_address = controller_address
        _logger.info(
            "Setting averaging period for rms current calculation of controller [%s] instrument [%s] to [%f]",
            self.controller_address, self._name, averaging_time)
        # instrument must be in CONFIGURATION state to set the current limit.
        with self.configuration_state():
            self._scpi_protocol.write(
                self._build_command("QIT", averaging_time))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

    @rpc_method
    def get_analog_input_value(self, controller_address: Optional[int] = None) -> float:
        """
        Get the analog input value.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            Analog input value in Volts.
        """
        self.controller_address = controller_address
        _logger.info("Getting analog input value of instrument [%s]", self._name)
        analog_input = self._scpi_protocol.ask(
            self._build_command("RA"))
        self._check_error()
        return float(analog_input[3:])

    @rpc_method
    def get_ttl_input_value(self, controller_address: Optional[int] = None) -> int:
        """
        Get the TTL input value. The returned decimal number represents the binary word made of all 4 inputs,
        where bit 0 is input 1, bit 1 is input 2, bit 2 is input 3, and bit 3 is input 4.

        The TTL input value is 1 when the corresponding voltage on the pin is larger than 2.4
        volts, and it is 0 when the corresponding voltage is below 0.8 volt. When the voltage is
        between these two values, the result is unreliable and can be 1 or 0.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            TTL input value in bits. E.g. '5' means input 0 and 2 are high, all others are low.
        """
        self.controller_address = controller_address
        _logger.info("Getting TTL input value of instrument [%s]", self._name)
        ttl_input = self._scpi_protocol.ask(
            self._build_command("RB"))
        self._check_error()
        return int(ttl_input[3:])

    @rpc_method
    def get_ttl_output_value(self, controller_address: Optional[int] = None) -> int:
        """
        Get the TTL output value. The returned decimal number represents the binary word made of all 4 outputs,
        where bit 0 is output 1, bit 1 is output 2, bit 2 is output 3, and bit 3 is output 4.

        A 1 represents closed collector output transistor of the output. A 0 represents open
        collector output transistor of the output.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            TTL output value in bits. E.g. '3' means TTL outputs 1 & 2 are closed and outputs 3 & 4 open.
        """
        self.controller_address = controller_address
        _logger.info("Getting TTL output value of instrument [%s]", self._name)
        ttl_output = self._scpi_protocol.ask(
            self._build_command("SB?"))
        self._check_error()
        return int(ttl_output[3:])

    @rpc_method
    def set_ttl_output_value(self, ttl_output_value: int, controller_address: Optional[int] = None) -> None:
        """
        Get the TTL output value. The value is a binary word made of all 4 outputs,
        where bit 0 is output 1, bit 1 is output 2, bit 2 is output 3, and bit 3 is output 4.

        A 1 closes the open collector output transistor of the output. A 0 blocks the open
        collector output transistor of the output.

        Parameters:
            ttl_output_value:   New TTL output value.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        if ttl_output_value not in range(16):
            raise QMI_InstrumentException(
                f"Provided value {ttl_output_value} not in valid range 0-15 (all open - all closed).")

        self.controller_address = controller_address
        _logger.info(
            "Setting the TTL output value for instrument [%s] to [%i]", self._name, ttl_output_value
        )
        self._scpi_protocol.write(self._build_command(
            "SB", ttl_output_value))
        self._check_error()

    @rpc_method
    def get_controller_rs485_address(self) -> int:
        """
        Get the controller’s RS-485 address. Controller address is always 1 for this command.

        Returns:
            rs485_address:      Controller's axis number for new RS485 address.
        """
        self.controller_address = self.DEFAULT_CONTROLLER_ADDRESS
        _logger.info(
            "Getting RS485 address of controller [%s] instrument [%s]", self.controller_address, self._name
        )
        # instrument must be in CONFIGURATION state to get the RS485 address.
        with self.configuration_state():
            axis = self._scpi_protocol.ask(
                self._build_command("SA?"))
            self._check_error()

        return int(axis[3:])

    @rpc_method
    def set_controller_rs485_address(self, rs485_address: int) -> None:
        """
        Set the controller’s RS-485 address. Controller address is always 1 for this command.

        Parameters:
            rs485_address:      Controller's axis number for new RS485 address.
        """
        if 2 > rs485_address or rs485_address > 31:
            raise QMI_InstrumentException(
                f"Invalid controller axis number {rs485_address}")

        self.controller_address = self.DEFAULT_CONTROLLER_ADDRESS
        _logger.info(
            "Setting RS485 address of controller [%s] instrument [%s] to axis [%i]",
            self.controller_address, self._name, rs485_address
        )
        # instrument must be in CONFIGURATION state to set the RS485 address.
        with self.configuration_state():
            self._scpi_protocol.write(
                self._build_command("SA", rs485_address))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

    @rpc_method
    def get_negative_software_limit(self, controller_address: Optional[int] = None) -> float:
        """
        Get the negative software limit.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            neg_sw_limit:      Controller's negative software limit.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting the negative software limit of controller [%s] instrument [%s]",
            self.controller_address, self._name
        )
        # instrument must be in CONFIGURATION, DISABLE or READY state to get the negative software limit.
        try:
            config_state = False
            self._state_ready_check("negative software limit")

        except QMI_InstrumentException:
            self._enter_configuration_state()
            config_state = True

        neg_sw_limit = self._scpi_protocol.ask(
            self._build_command("SL?"))
        self._check_error()
        if config_state:
            self._exit_configuration_state()

        return float(neg_sw_limit[3:])

    @rpc_method
    def set_negative_software_limit(
            self, neg_sw_limit: float, persist: bool = False, controller_address: Optional[int] = None
    ) -> None:
        """
        Set the negative software limit.

        Parameters:
            neg_sw_limit:       Controller's negative software limit.
            persist:            Flag to indicate if the software limit should be persisted to the controller's memory,
                                so it is still available after powering down the controller. When not persisted, the
                                negative software limit is the one stored in the controller's memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        def _set_negative_software_limit():
            self._scpi_protocol.write(
                self._build_command("SL", neg_sw_limit))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

        if neg_sw_limit <= -self.MAX_FLOAT_LIMIT or neg_sw_limit > 0:
            raise QMI_InstrumentException(
                f"Negative software limit {neg_sw_limit} not in valid range -{self.MAX_FLOAT_LIMIT} "
                "< neg_sw_limit <= 0"
            )

        self.controller_address = controller_address
        _logger.info(
            "Setting negative software limit of controller [%s] instrument [%s] to [%f]",
            self.controller_address, self._name, neg_sw_limit)
        if persist:
            # instrument must be in CONFIGURATION state to set the software limit.
            with self.configuration_state():
                _set_negative_software_limit()

        else:
            # instrument must be in DISABLE or READY state to set the software limit.
            self._state_ready_check("negative software limit")
            _set_negative_software_limit()

    @rpc_method
    def get_positive_software_limit(self, controller_address: Optional[int] = None) -> float:
        """
        Get the positive software limit.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            pos_sw_limit:      Controller's positive software limit.
        """
        def _get_positive_software_limit() -> str:
            pos_sw_limit = self._scpi_protocol.ask(
                self._build_command("SR?"))
            self._check_error()
            return pos_sw_limit

        self.controller_address = controller_address
        _logger.info(
            "Getting the positive software limit of controller [%s] instrument [%s]",
            self.controller_address, self._name
        )
        # instrument must be in CONFIGURATION, DISABLE or READY state to get the positive software limit.
        try:
            self._state_ready_check("positive software limit")
            pos_sw_limit = _get_positive_software_limit()

        except QMI_InstrumentException:
            with self.configuration_state():
                pos_sw_limit = _get_positive_software_limit()

        return float(pos_sw_limit[3:])

    @rpc_method
    def set_positive_software_limit(
            self, pos_sw_limit: float, persist: bool = False, controller_address: Optional[int] = None) -> None:
        """
        Set the positive software limit.

        Parameters:
            pos_sw_limit:       Controller's positive software limit.
            persist:            Flag to indicate if the software limit should be persisted to the controller's memory,
                                so it is still available after powering down the controller. When not persisted, the
                                positive software limit is the one stored in the controller's memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        def _set_positive_software_limit():
            self._scpi_protocol.write(
                self._build_command("SR", pos_sw_limit))
            sleep(self.COMMAND_EXEC_TIME)
            self._check_error()

        if pos_sw_limit < 0 or pos_sw_limit >= self.MAX_FLOAT_LIMIT:
            raise QMI_InstrumentException(
                f"Positive software limit {pos_sw_limit} not in valid range 0 <= "
                f"pos_sw_limit <= {self.MAX_FLOAT_LIMIT}"
            )

        self.controller_address = controller_address
        _logger.info(
            "Setting positive software limit of controller [%s] instrument [%s] to [%f]",
            self.controller_address, self._name, pos_sw_limit)
        if persist:
            # instrument must be in CONFIGURATION state to set the software limit.
            with self.configuration_state():
                _set_positive_software_limit()

        else:
            # instrument must be in DISABLE or READY state to set the software limit.
            self._state_ready_check("positive software limit")
            _set_positive_software_limit()



"""
Instrument driver for the Newport CONEX-CC DC servo motion controller.
"""
from typing import Dict, Optional
from qmi.core.context import QMI_Context
from qmi.instruments.newport.actuators import LinearActuator
from qmi.instruments.newport.single_axis_motion_controller import Newport_SingleAxisMotionController


class Newport_ConexCC(Newport_SingleAxisMotionController):
    """Instrument driver for the Newport CONEX-CC DC servo motion controller."""

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str,
                 serial: str,
                 actuators: Dict[Optional[int], LinearActuator],
                 baudrate: int = 921600) -> None:
        """Initialize driver.

        Parameters:
            name:       Name for this instrument instance.
            transport:  QMI transport descriptor to connect to the instrument.
            serial:     The serial number of the device.
            actuators:  The linear actuator that this controller will drive.
            baudrate:   The baudrate of the instrument. Defaults to 921600.
        """
        super().__init__(context, name, transport, serial, actuators, baudrate)



#! /usr/bin/env python3
"""Unit-tests for testing context managing for `QMI_Instrument` class."""
import logging
import unittest

from numpy import random

from tests.patcher import PatcherQmiContext as QMI_Context
from qmi.core.context import QMI_Instrument, rpc_method


class MyInstrument_TestDriver(QMI_Instrument):

    def __init__(self, context: QMI_Context, name: str) -> None:
        super().__init__(context, name)
        self._it_is_open = False

    @rpc_method
    def open(self):
        super().open()
        self._it_is_open = True

    @rpc_method
    def close(self):
        super().close()
        self._it_is_open = False

    @rpc_method
    def is_it_open_then(self):
        return self._it_is_open


class TestInstrumentContextManager(unittest.TestCase):

    def setUp(self):
        # Suppress logging.
        logging.getLogger("qmi.core.rpc").setLevel(logging.CRITICAL)

        self._ctx_qmi_id = f"test-qmi-instrument-{random.randint(0, 100)}"
        self.qmi_patcher = QMI_Context("test_instrument_context_manager")
        self.qmi_patcher.start(self._ctx_qmi_id)

    def tearDown(self) -> None:
        self.qmi_patcher.stop()

    def test_create_instance(self):
        """Test creating an instrument object works without starting qmi."""
        instr = MyInstrument_TestDriver(self.qmi_patcher, self._ctx_qmi_id)
        # Get RPC object info
        name = instr.get_name()
        category = instr.get_category()
        # Assert
        self.assertFalse(instr.is_open())
        self.assertEqual(self._ctx_qmi_id, name)
        self.assertEqual("instrument", category)

    def test_make_instrument_with_with(self):
        """Test creating an instrument object with context manager and also opens and closes it."""
        with MyInstrument_TestDriver(self.qmi_patcher, self._ctx_qmi_id) as instr:
            # Assert
            name = instr.get_name()
            category = instr.get_category()
            # Assert
            self.assertTrue(instr.is_open())
            self.assertEqual(self._ctx_qmi_id, name)
            self.assertEqual("instrument", category)
            # Also see extra action in the instrument driver's `open` has been executed
            self.assertTrue(instr.is_it_open_then())

        # Assert
        self.assertFalse(instr.is_open())
        # Also see extra action in the instrument driver's `close` has been executed
        self.assertFalse(instr.is_it_open_then())


class TestInstrumentMakeFunction(unittest.TestCase):

    def setUp(self):
        # Suppress logging.
        logging.getLogger("qmi.core.rpc").setLevel(logging.CRITICAL)

        # Start a context with creating the "instrument".
        self.c1 = QMI_Context("c1")
        self.c1.start()

    def tearDown(self):
        self.c1.stop()
        self.c1 = None

        logging.getLogger("qmi.core.rpc").setLevel(logging.NOTSET)

    def test_make_instrument(self):
        """Test making the instrument with 'make_instrument' works as expected"""
        instr_proxy = self.c1.make_instrument("instr", MyInstrument_TestDriver)

        # Assert
        self.assertFalse(instr_proxy.is_open())

    def test_make_instrument_with_with(self):
        """Test making the instrument with 'make_instrument' works also with context manager."""
        with self.c1.make_instrument("instr", MyInstrument_TestDriver) as instr_proxy:
            # Assert
            self.assertTrue(instr_proxy.is_open())
            # Also see extra action in the instrument driver's `open` has been executed
            self.assertTrue(instr_proxy.is_it_open_then())

        # Assert
        self.assertFalse(instr_proxy.is_open())
        # Also see extra action in the instrument driver's `close` has been executed
        self.assertFalse(instr_proxy.is_it_open_then())


if __name__ == '__main__':
    unittest.main()



"""

Python USBTMC driver

Copyright (c) 2012-2017 Alex Forencich

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

====

This module was modified for use in QMI.

Original code downloaded from https://github.com/python-ivi/python-usbtmc/commit d9bfb20b2ef002da787adb6b093e1679705c00e2
Original author: Alex Forencich
Original license: MIT License

This module was added to the QMI repository because the original repository
is no longer maintained and has unresolved issues.

Modifications by Joris van Rantwijk:
 (see also https://github.com/jorisvr/python-usbtmc/):

 - Read Bulk-IN data during abort_bulk_in sequence.
 - Clear Bulk-OUT endpoint after abort_bulk_out.
 - Ignore EOM flag when device sends partial data.
 - Always call set_configuration() to synchronize data toggle.
 - Clear Bulk-IN endpoint during initialization to synchronize data toggle.
"""

import usb.core
import usb.util
import struct
import time
import os
import re
import sys

# constants
USBTMC_bInterfaceClass    = 0xFE
USBTMC_bInterfaceSubClass = 3
USBTMC_bInterfaceProtocol = 0
USB488_bInterfaceProtocol = 1

USBTMC_MSGID_DEV_DEP_MSG_OUT            = 1
USBTMC_MSGID_REQUEST_DEV_DEP_MSG_IN     = 2
USBTMC_MSGID_DEV_DEP_MSG_IN             = 2
USBTMC_MSGID_VENDOR_SPECIFIC_OUT        = 126
USBTMC_MSGID_REQUEST_VENDOR_SPECIFIC_IN = 127
USBTMC_MSGID_VENDOR_SPECIFIC_IN         = 127
USB488_MSGID_TRIGGER                    = 128

USBTMC_STATUS_SUCCESS                  = 0x01
USBTMC_STATUS_PENDING                  = 0x02
USBTMC_STATUS_FAILED                   = 0x80
USBTMC_STATUS_TRANSFER_NOT_IN_PROGRESS = 0x81
USBTMC_STATUS_SPLIT_NOT_IN_PROGRESS    = 0x82
USBTMC_STATUS_SPLIT_IN_PROGRESS        = 0x83
USB488_STATUS_INTERRUPT_IN_BUSY        = 0x20

USBTMC_REQUEST_INITIATE_ABORT_BULK_OUT     = 1
USBTMC_REQUEST_CHECK_ABORT_BULK_OUT_STATUS = 2
USBTMC_REQUEST_INITIATE_ABORT_BULK_IN      = 3
USBTMC_REQUEST_CHECK_ABORT_BULK_IN_STATUS  = 4
USBTMC_REQUEST_INITIATE_CLEAR              = 5
USBTMC_REQUEST_CHECK_CLEAR_STATUS          = 6
USBTMC_REQUEST_GET_CAPABILITIES            = 7
USBTMC_REQUEST_INDICATOR_PULSE             = 64

USB488_READ_STATUS_BYTE = 128
USB488_REN_CONTROL      = 160
USB488_GOTO_LOCAL       = 161
USB488_LOCAL_LOCKOUT    = 162

USBTMC_HEADER_SIZE = 12

RIGOL_QUIRK_PIDS = [0x04ce, 0x0588]


def parse_visa_resource_string(resource_string):
    """Parse VISA resource string."""
    # valid resource strings:
    # USB::1234::5678::INSTR
    # USB::1234::5678::SERIAL::INSTR
    # USB0::0x1234::0x5678::INSTR
    # USB0::0x1234::0x5678::SERIAL::INSTR
    m = re.match(r'^(?P<prefix>(?P<type>USB)\d*)(::(?P<arg1>[^\s:]+))'
                 r'(::(?P<arg2>[^\s:]+(\[.+\])?))(::(?P<arg3>[^\s:]+))?'
                 r'(::(?P<suffix>INSTR))$', resource_string, re.I)

    if m is not None:
        return dict(
            type=m.group('type').upper(),
            prefix=m.group('prefix'),
            arg1=m.group('arg1'),
            arg2=m.group('arg2'),
            arg3=m.group('arg3'),
            suffix=m.group('suffix')
        )


# Exceptions
class UsbtmcException(Exception):
    """USBTMC exception type."""
    em = {0: "No error"}

    def __init__(self, err=None, note=None):
        self.err = err
        self.note = note
        self.msg = ''

        # Compose error message
        if err is None:
            self.msg = note
        else:
            if type(err) is int:
                if err in self.em:
                    self.msg = "%d: %s" % (err, self.em[err])
                else:
                    self.msg = "%d: Unknown error" % err
            else:
                self.msg = err
            if note is not None:
                self.msg = "%s [%s]" % (self.msg, note)

    def __str__(self):
        return self.msg


def list_devices():
    """List all connected USBTMC devices"""

    # Filter function.
    def is_usbtmc_device(dev):
        for cfg in dev:
            d = usb.util.find_descriptor(cfg, bInterfaceClass=USBTMC_bInterfaceClass,
                                         bInterfaceSubClass=USBTMC_bInterfaceSubClass)
            if d is not None:
                return True

            if dev.idVendor == 0x1334:
                # Advantest
                return True

            if dev.idVendor == 0x0957:
                # Agilent
                if dev.idProduct in [0x2818, 0x4218, 0x4418]:
                    # Agilent U27xx modular devices in firmware update mode
                    # 0x2818 for U2701A/U2702A (firmware update mode on power up)
                    # 0x4218 for U2722A (firmware update mode on power up)
                    # 0x4418 for U2723A (firmware update mode on power up)
                    return True

        return False

    try:
        return list(usb.core.find(find_all=True, custom_match=is_usbtmc_device))
    except usb.core.USBError:
        # No devices present
        return []


def list_resources():
    """List resource strings for all connected USBTMC devices"""

    res = []

    for dev in list_devices():
        idVendor = dev.idVendor
        idProduct = dev.idProduct

        # "fix" IDs for devices in firmware update mode
        if idVendor == 0x0957 and idProduct == 0x2818:
            # Agilent U2701A/U2702A firmware update mode
            idProduct = 0x2918

        if idVendor == 0x0957 and idProduct == 0x4218:
            # Agilent U2722A firmware update mode
            idProduct = 0x4118

        if idVendor == 0x0957 and idProduct == 0x4418:
            # Agilent U2723A firmware update mode
            idProduct = 0x4318

        # Attempt to read serial number.
        iSerial = None
        try:
            iSerial = dev.serial_number
        except:
            # Maybe not set or user does not have read privilege.
            pass

        # Append formatted resource string to list.
        if iSerial is None:
            res.append("USB::%d::%d::INSTR" % (idVendor, idProduct))
        else:
            res.append("USB::%d::%d::%s::INSTR" % (idVendor, idProduct, iSerial))

    return res


def find_device(idVendor=None, idProduct=None, iSerial=None):
    "Find USBTMC instrument"

    devs = list_devices()

    if len(devs) == 0:
        return None

    for dev in devs:
        # Match VID and PID
        found = dev.idVendor == idVendor and dev.idProduct == idProduct

        if idVendor == 0x0957 and idProduct == 0x2918:
            # Agilent U2701A/U2702A firmware update mode
            if dev.idVendor == idVendor and dev.idProduct == 0x2818:
                found = True

        if idVendor == 0x0957 and idProduct == 0x4118:
            # Agilent U2722A firmware update mode
            if dev.idVendor == idVendor and dev.idProduct == 0x4218:
                found = True

        if idVendor == 0x0957 and idProduct == 0x4318:
            # Agilent U2723A firmware update mode
            if dev.idVendor == idVendor and dev.idProduct == 0x4418:
                found = True

        if not found:
            continue

        if iSerial is None:
            return dev
        else:
            s = ''

            # Try reading serial number
            try:
                s = dev.serial_number
            except:
                pass

            if iSerial == s:
                return dev

    return None


class Instrument(object):
    "USBTMC instrument interface client"
    def __init__(self, *args, **kwargs):
        """Create new USBTMC instrument object."""
        self.idVendor = 0
        self.idProduct = 0
        self.iSerial = None
        self.device = None
        self.cfg = None
        self.iface = None
        self.term_char = None

        self.bcdUSBTMC = 0
        self.support_pulse = False
        self.support_talk_only = False
        self.support_listen_only = False
        self.support_term_char = False

        self.bcdUSB488 = 0
        self.support_USB4882 = False
        self.support_remote_local = False
        self.support_trigger = False
        self.support_scpi = False
        self.support_SR = False
        self.support_RL = False
        self.support_DT = False

        self.max_transfer_size = 1024*1024

        self.timeout = 5.0

        # Separate timeout for abort sequences. It is reasonable to wait
        # longer during an abort sequence, to avoid leaving the device
        # in an unusable state.
        self.abort_timeout = 5.0

        self.bulk_in_ep = None
        self.bulk_out_ep = None
        self.interrupt_in_ep = None

        self.last_btag = 0
        self.last_rstb_btag = 0

        self.connected = False
        self.reattach = []
        self.old_cfg = None

        # quirks
        self.advantest_quirk = False
        self.advantest_locked = False

        self.rigol_quirk = False
        self.rigol_quirk_ieee_block = False

        # Enable always calling set_configuration() while opening the device,
        # even if the correct configuration was already active. This is useful
        # because set_configuration() resets the DATA0/DATA1 toggle state.
        # If the data toggle state becomes mismatched between host and device,
        # data will be lost, typically causing timeouts during the first
        # interaction with the device.
        self.force_reconfigure = True

        # Enable clearing Bulk-IN endpoint halt during initialization.
        # This is not strictly necessary, but it helps to reset the DATA0/DATA1
        # toggle state for some devices which do not correctly reset
        # the toggle state during set_configuration().
        self.force_clear_bulk_in = True

        resource = None

        # Process arguments
        if len(args) == 1:
            if type(args[0]) == str:
                resource = args[0]
            else:
                self.device = args[0]
        if len(args) >= 2:
            self.idVendor = args[0]
            self.idProduct = args[1]
        if len(args) >= 3:
            self.iSerial = args[2]

        # Process keyword arguments
        for op, val in kwargs.items():
            if op == 'idVendor':
                self.idVendor = val
            elif op == 'idProduct':
                self.idProduct = val
            elif op == 'iSerial':
                self.iSerial = val
            elif op == 'device':
                self.device = val
            elif op == 'dev':
                self.device = val
            elif op == 'term_char':
                self.term_char = val
            elif op == 'resource':
                resource = val

        # Parse resource string, if given.
        if resource is not None:
            res = parse_visa_resource_string(resource)

            if res is None:
                raise UsbtmcException("Invalid resource string", 'init')

            if res['arg1'] is None and res['arg2'] is None:
                raise UsbtmcException("Invalid resource string", 'init')

            self.idVendor = int(res['arg1'], 0)
            self.idProduct = int(res['arg2'], 0)
            self.iSerial = res['arg3']

        # Find device
        if self.device is None:
            if self.idVendor is None or self.idProduct is None:
                raise UsbtmcException("No device specified", 'init')
            else:
                self.device = find_device(self.idVendor, self.idProduct, self.iSerial)
                if self.device is None:
                    raise UsbtmcException("Device not found", 'init')

    def __del__(self):
        if self.connected:
            self.close()

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, val):
        self._timeout = val
        self._timeout_ms = int(val * 1000)

    def open(self):
        if self.connected:
            return

        # Initialize device
        if self.device.idVendor == 0x0957 and self.device.idProduct in [0x2818, 0x4218, 0x4418]:
            self._init_agilent_u27xx()

        # find first USBTMC interface
        for cfg in self.device:
            for iface in cfg:
                if (iface.bInterfaceClass == USBTMC_bInterfaceClass and
                        iface.bInterfaceSubClass == USBTMC_bInterfaceSubClass):
                    # USBTMC device
                    self.cfg = cfg
                    self.iface = iface
                    break
                elif (self.device.idVendor == 0x1334):
                    # Advantest
                    self.cfg = cfg
                    self.iface = iface
                    break
                else:
                    continue
            break

        if self.iface is None:
            raise UsbtmcException("Not a USBTMC device", 'init')

        try:
            self.old_cfg = self.device.get_active_configuration()
        except usb.core.USBError:
            # ignore exception if configuration is not set
            pass

        if (self.old_cfg is not None
                and self.old_cfg.bConfigurationValue == self.cfg.bConfigurationValue
                and not self.force_reconfigure):
            # already set to correct configuration

            # release kernel driver on USBTMC interface
            self._release_kernel_driver(self.iface.bInterfaceNumber)
        else:
            # wrong configuration or configuration not set

            # release all kernel drivers
            if self.old_cfg is not None:
                for iface in self.old_cfg:
                    self._release_kernel_driver(iface.bInterfaceNumber)

            # set proper configuration
            self.device.set_configuration(self.cfg)

        # claim interface
        usb.util.claim_interface(self.device, self.iface)

        # don't need to set altsetting - USBTMC devices have 1 altsetting as per the spec

        # find endpoints
        for ep in self.iface:
            ep_dir = usb.util.endpoint_direction(ep.bEndpointAddress)
            ep_type = usb.util.endpoint_type(ep.bmAttributes)

            if (ep_type == usb.util.ENDPOINT_TYPE_BULK):
                if (ep_dir == usb.util.ENDPOINT_IN):
                    self.bulk_in_ep = ep
                elif (ep_dir == usb.util.ENDPOINT_OUT):
                    self.bulk_out_ep = ep
            elif (ep_type == usb.util.ENDPOINT_TYPE_INTR):
                if (ep_dir == usb.util.ENDPOINT_IN):
                    self.interrupt_in_ep = ep

        if self.bulk_in_ep is None or self.bulk_out_ep is None:
            raise UsbtmcException("Invalid endpoint configuration", 'init')

        self._handle_vendor_quirks()

        self.connected = True

        self.clear()

        self.get_capabilities()

    def _init_agilent_u27xx(self):
        """Handle init of  Agilent U27xx modular devices U2701A/U2702A, U2722A/U2723A

        These devices require a short initialization sequence, presumably
        to take them out of 'firmware update' mode after confirming
        that the firmware version is correct. This is required once
        on every power-on before the device can be used.
        Note that the device will reset and the product ID will change.
        U2701A/U2702A boot 0x2818, usbtmc 0x2918
        U2722A boot 0x4218, usbtmc 0x4118
        U2723A boot 0x4418, usbtmc 0x4318
        """
        serial = self.device.serial_number
        new_id = 0

        if self.device.idProduct == 0x2818:
            # U2701A/U2702A
            new_id = 0x2918
            self.device.ctrl_transfer(bmRequestType=0xC0, bRequest=0x0C, wValue=0x0000, wIndex=0x047E, data_or_wLength=0x0001)
            self.device.ctrl_transfer(bmRequestType=0xC0, bRequest=0x0C, wValue=0x0000, wIndex=0x047D, data_or_wLength=0x0006)
            self.device.ctrl_transfer(bmRequestType=0xC0, bRequest=0x0C, wValue=0x0000, wIndex=0x0484, data_or_wLength=0x0005)
            self.device.ctrl_transfer(bmRequestType=0xC0, bRequest=0x0C, wValue=0x0000, wIndex=0x0472, data_or_wLength=0x000C)
            self.device.ctrl_transfer(bmRequestType=0xC0, bRequest=0x0C, wValue=0x0000, wIndex=0x047A, data_or_wLength=0x0001)
            self.device.ctrl_transfer(bmRequestType=0x40, bRequest=0x0C, wValue=0x0000, wIndex=0x0475, data_or_wLength=b'\x00\x00\x01\x01\x00\x00\x08\x01')

        if self.device.idProduct in [0x4218, 0x4418]:
            # U2722A/U2723A
            if self.device.idProduct == 0x4218:
                # U2722A
                new_id = 0x4118
            elif self.device.idProduct == 0x4418:
                # U2723A
                new_id = 0x4318
            self.device.ctrl_transfer(bmRequestType=0xC0, bRequest=0x0C, wValue=0x0000, wIndex=0x047E, data_or_wLength=0x0001)
            self.device.ctrl_transfer(bmRequestType=0xC0, bRequest=0x0C, wValue=0x0000, wIndex=0x047D, data_or_wLength=0x0006)
            self.device.ctrl_transfer(bmRequestType=0xC0, bRequest=0x0C, wValue=0x0000, wIndex=0x0487, data_or_wLength=0x0005)
            self.device.ctrl_transfer(bmRequestType=0xC0, bRequest=0x0C, wValue=0x0000, wIndex=0x0472, data_or_wLength=0x000C)
            self.device.ctrl_transfer(bmRequestType=0xC0, bRequest=0x0C, wValue=0x0000, wIndex=0x047A, data_or_wLength=0x0001)
            self.device.ctrl_transfer(bmRequestType=0x40, bRequest=0x0C, wValue=0x0000, wIndex=0x0475, data_or_wLength=b'\x00\x00\x01\x01\x00\x00\x08\x01')

        usb.util.dispose_resources(self.device)
        self.device = None

        for i in range(40):
            self.device = find_device(0x0957, new_id, serial)
            if self.device is not None:
                break
            time.sleep(0.5)

        if self.device is None:
            raise RuntimeError("Agilent U27xx modular device initialization failed")

    def _handle_vendor_quirks(self):
        """Handle vendor/device specific quirks."""
        # Set quirk flags if necessary
        if self.device.idVendor == 0x1334:
            # Advantest/ADCMT devices have a very odd USBTMC implementation
            # which requires max 63 byte reads and never signals EOI on read
            self.max_transfer_size = 63
            self.advantest_quirk = True

        if self.device.idVendor == 0x1ab1 and self.device.idProduct in RIGOL_QUIRK_PIDS:
            # Rigol.
            self.rigol_quirk = True

            if self.device.idProduct == 0x04ce:
                self.rigol_quirk_ieee_block = True

        if self.device.idVendor == 0x1313:
            # Thorlabs devices don't respond otherwise (code provided by vendor).
            self.device.ctrl_transfer(bmRequestType=0xA1, bRequest=USB488_REN_CONTROL, wValue=0x0001, wIndex=0x0000, data_or_wLength=1)

    def close(self):
        """Close device."""
        if not self.connected:
            return

        if self.device.idVendor == 0x1313:
            # Thorlabs-specific cleanup (code provided by vendor).
            self.device.ctrl_transfer(bmRequestType=0xA1, bRequest=USB488_GOTO_LOCAL, wValue=0x0000, wIndex=0x0000, data_or_wLength=1)
            self.device.ctrl_transfer(bmRequestType=0xA1, bRequest=USB488_REN_CONTROL, wValue=0x0000, wIndex=0x0000, data_or_wLength=1)
            time.sleep(0.1)

        usb.util.dispose_resources(self.device)

        try:
            # Reset configuration.
            if self.cfg.bConfigurationValue != self.old_cfg.bConfigurationValue:
                self.device.set_configuration(self.old_cfg)

            # Try to reattach kernel driver
            for iface in self.reattach:
                try:
                    self.device.attach_kernel_driver(iface)
                except:
                    pass
        except:
            pass

        self.reattach = []
        self.connected = False

    def is_usb488(self):
        """Check if device supports USB488."""
        return self.iface.bInterfaceProtocol == USB488_bInterfaceProtocol

    def get_capabilities(self):
        """Enumerate the capabilities of the device."""

        # Open device if needed.
        if not self.connected:
            self.open()

        # Request capabilities.
        b = self.device.ctrl_transfer(
              usb.util.build_request_type(usb.util.CTRL_IN, usb.util.CTRL_TYPE_CLASS, usb.util.CTRL_RECIPIENT_INTERFACE),
              USBTMC_REQUEST_GET_CAPABILITIES,
              0x0000,
              self.iface.index,
              0x0018,
              timeout=self._timeout_ms)

        # Parse return value.
        if (b[0] == USBTMC_STATUS_SUCCESS):
            self.bcdUSBTMC = (b[3] << 8) + b[2]
            self.support_pulse = b[4] & 4 != 0
            self.support_talk_only = b[4] & 2 != 0
            self.support_listen_only = b[4] & 1 != 0
            self.support_term_char = b[5] & 1 != 0

            if self.is_usb488():
                self.bcdUSB488 = (b[13] << 8) + b[12]
                self.support_USB4882 = b[4] & 4 != 0
                self.support_remote_local = b[4] & 2 != 0
                self.support_trigger = b[4] & 1 != 0
                self.support_scpi = b[4] & 8 != 0
                self.support_SR = b[4] & 4 != 0
                self.support_RL = b[4] & 2 != 0
                self.support_DT = b[4] & 1 != 0
        else:
            raise UsbtmcException("Get capabilities failed", 'get_capabilities')

    def pulse(self):
        """Send a pulse indicator request, this should blink a light for 500-1000ms and then turn off again.

        Only for devices that support this.
        """
        # Open device if needed.
        if not self.connected:
            self.open()

        if self.support_pulse:
            # Send pulse command.
            b = self.device.ctrl_transfer(
                  usb.util.build_request_type(usb.util.CTRL_IN, usb.util.CTRL_TYPE_CLASS, usb.util.CTRL_RECIPIENT_INTERFACE),
                  USBTMC_REQUEST_INDICATOR_PULSE,
                  0x0000,
                  self.iface.index,
                  0x0001,
                  timeout=self._timeout_ms)

            # Check returned result.
            if (b[0] != USBTMC_STATUS_SUCCESS):
                raise UsbtmcException("Pulse failed", 'pulse')

    def pack_bulk_out_header(self, msgid):
        # Update bulk tag.
        self.last_btag = btag = (self.last_btag % 255) + 1

        # Create header.
        return struct.pack('BBBx', msgid, btag, ~btag & 0xFF)

    def pack_dev_dep_msg_out_header(self, transfer_size, eom = True):
        hdr = self.pack_bulk_out_header(USBTMC_MSGID_DEV_DEP_MSG_OUT)
        return hdr+struct.pack("<LBxxx", transfer_size, eom)

    def pack_dev_dep_msg_in_header(self, transfer_size, term_char = None):
        hdr = self.pack_bulk_out_header(USBTMC_MSGID_DEV_DEP_MSG_IN)
        transfer_attributes = 0
        if term_char is None:
            term_char = 0
        else:
            transfer_attributes = 2
            term_char = self.term_char
        return hdr+struct.pack("<LBBxx", transfer_size, transfer_attributes, term_char)

    def pack_vendor_specific_out_header(self, transfer_size):
        hdr = self.pack_bulk_out_header(USBTMC_MSGID_VENDOR_SPECIFIC_OUT)
        return hdr+struct.pack("<Lxxxx", transfer_size)

    def pack_vendor_specific_in_header(self, transfer_size):
        hdr = self.pack_bulk_out_header(USBTMC_MSGID_VENDOR_SPECIFIC_IN)
        return hdr+struct.pack("<Lxxxx", transfer_size)

    def pack_usb488_trigger(self):
        hdr = self.pack_bulk_out_header(USB488_MSGID_TRIGGER)
        return hdr+b'\x00'*8

    def unpack_bulk_in_header(self, data):
        msgid, btag, btaginverse = struct.unpack_from('BBBx', data)
        return (msgid, btag, btaginverse)

    def unpack_dev_dep_resp_header(self, data):
        msgid, btag, btaginverse = self.unpack_bulk_in_header(data)
        transfer_size, transfer_attributes = struct.unpack_from('<LBxxx', data, 4)
        data = data[USBTMC_HEADER_SIZE:transfer_size+USBTMC_HEADER_SIZE]
        return (msgid, btag, btaginverse, transfer_size, transfer_attributes, data)

    def write_raw(self, data):
        """Write binary data to instrument

        Parameters:
            data:   data to send.
        """
        # Open device if needed.
        if not self.connected:
            self.open()

        # Send data.
        eom = False
        num = len(data)
        offset = 0
        try:
            while num > 0:
                # Check if remaining data fits in a single message.
                if num <= self.max_transfer_size:
                    eom = True

                # Create block of data for current message.
                block = data[offset:offset+self.max_transfer_size]
                size = len(block)

                # Send message
                req = self.pack_dev_dep_msg_out_header(size, eom) + block + b'\0'*((4 - (size % 4)) % 4)
                self.bulk_out_ep.write(req, timeout=self._timeout_ms)

                # Update for next block.
                offset += size
                num -= size
        except usb.core.USBError:
            exc = sys.exc_info()[1]
            if exc.errno == 110:
                # Timeout, abort transfer.
                self._abort_bulk_out()
            raise

    def read_raw(self, num=-1):
        """Read binary data from instrument

        Parameters:
            num:    number of bytes to read, or -1 for everything that is available.
        """
        # Open device if needed.
        if not self.connected:
            self.open()

        # Check if requested number of bytes fit in a single message.
        read_len = self.max_transfer_size
        if 0 < num < read_len:
            read_len = num

        # Determine what message terminator to use.
        term_char = None
        if self.term_char is not None:
            term_char = self.term_char

        # Read data.
        eom = False
        read_data = b''
        try:
            while not eom:
                if not self.rigol_quirk or read_data == b'':
                    # If the rigol sees this again, it will restart the transfer so only send it the first time.
                    req = self.pack_dev_dep_msg_in_header(read_len, term_char)
                    self.bulk_out_ep.write(req, timeout=self._timeout_ms)

                # Read a single message.
                resp = self.bulk_in_ep.read(read_len+USBTMC_HEADER_SIZE+3, timeout=self._timeout_ms)

                # Handle Python byte array stuff.
                if sys.version_info >= (3, 2):
                    resp = resp.tobytes()
                else:
                    resp = resp.tostring()

                # Parse header.
                if self.rigol_quirk and read_data:
                    pass # do nothing, the packet has no header if it isn't the first
                else:
                    msgid, btag, btaginverse, transfer_size, transfer_attributes, data = \
                        self.unpack_dev_dep_resp_header(resp)

                if self.rigol_quirk:
                    # RIGOL devices only send the header in the first packet, and they lie about whether the
                    # transaction is complete
                    if read_data:
                        read_data += resp
                    else:
                        if self.rigol_quirk_ieee_block and data.startswith(b"#"):
                            # ieee block incoming, the transfer_size usbtmc header is lying about the transaction size
                            l = int(chr(data[1]))
                            n = int(data[2:l+2])
                            transfer_size = n + (l+2)  # account for ieee header

                        read_data += data

                    # Check if there is more data to expect.
                    if len(read_data) >= transfer_size:
                        read_data = read_data[:transfer_size]  # as per usbtmc spec section 3.2 note 2
                        eom = True
                    else:
                        eom = False
                else:
                    # Only consider EOM flag when transfer_size bytes received.
                    # See USBTMC v1.00 3.3.1.1:
                    #     "The host must ignore EOM if the device does not
                    #     send TransferSize message data bytes."
                    if len(data) >= transfer_size:
                        eom = (transfer_attributes & 1) != 0
                    else:
                        eom = False
                    read_data += data

                # Advantest devices never signal EOI and may only send one read packet
                if self.advantest_quirk:
                    break

                # Check if more data is expected.
                if num > 0:
                    num = num - len(data)
                    if num <= 0:
                        break
                    if num < read_len:
                        read_len = num
        except usb.core.USBError:
            exc = sys.exc_info()[1]
            if exc.errno == 110:
                # Timeout, abort transfer.
                self._abort_bulk_in()
            raise

        return read_data

    def ask_raw(self, data, num=-1):
        """Write (ask) then read binary data.

        Parameters:
            data:   data to send (ask).
            num:    number of response bytes to read, or -1 for everything (default: -1).
        """
        # Advantest/ADCMT hardware won't respond to a command unless it's in Local Lockout mode
        was_locked = self.advantest_locked

        try:
            # Lock Advantest devices.
            if self.advantest_quirk and not was_locked:
                self.lock()

            # Ask and get response.
            self.write_raw(data)
            return self.read_raw(num)
        finally:
            # Unlock Advantest devices.
            if self.advantest_quirk and not was_locked:
                self.unlock()

    def write(self, message, encoding='utf-8'):
        """Write string message(s) to instrument.

        Parameters:
            message:    single message as string, or list/tuple of messages.
            encoding:   encoding to use (default: utf-8).
        """
        # Recursive call for a list of commands.
        if type(message) is tuple or type(message) is list:
            for message_i in message:
                self.write(message_i, encoding)
            return

        # Write the message as raw string.
        self.write_raw(str(message).encode(encoding))

    def read(self, num=-1, encoding='utf-8'):
        """Read string message from instrument.

        Parameters:
            num:        number of bytes to read, or -1 for everything (default: -1).
            encoding:   encoding to use to decode the raw data (default: utf-8).
        """
        return self.read_raw(num).decode(encoding).rstrip('\r\n')

    def ask(self, message, num=-1, encoding='utf-8'):
        """Write (ask) then read string messages.

        Parameters:
            message:    message to send (ask).
            num:        number of response bytes to read, or -1 for everything (default: -1).
            encoding:   encoding to use to encode/decode the raw data (default: utf-8).
        """
        # Recursive call for a list of commands.
        if type(message) is tuple or type(message) is list:
            val = list()
            for message_i in message:
                val.append(self.ask(message_i, num, encoding))
            return val

        # Advantest/ADCMT hardware won't respond to a command unless it's in Local Lockout mode
        was_locked = self.advantest_locked

        try:
            # Lock Advantest devices.
            if self.advantest_quirk and not was_locked:
                self.lock()

            # Ask and get response.
            self.write(message, encoding)
            return self.read(num, encoding)
        finally:
            # Unlock Advantest devices.
            if self.advantest_quirk and not was_locked:
                self.unlock()

    def read_stb(self):
        """Read status byte."""
        # Open device if needed.
        if not self.connected:
            self.open()

        # Check which protocol to use.
        if self.is_usb488():
            # Increment status byte request tag.
            rstb_btag = (self.last_rstb_btag % 128) + 1
            if rstb_btag < 2:
                rstb_btag = 2
            self.last_rstb_btag = rstb_btag

            # Request status byte.
            b = self.device.ctrl_transfer(
                bmRequestType=usb.util.build_request_type(usb.util.CTRL_IN, usb.util.CTRL_TYPE_CLASS, usb.util.CTRL_RECIPIENT_INTERFACE),
                bRequest=USB488_READ_STATUS_BYTE,
                wValue=rstb_btag,
                wIndex=self.iface.index,
                data_or_wLength=0x0003,
                timeout=self._timeout_ms
            )

            # Parse response.
            if (b[0] == USBTMC_STATUS_SUCCESS):
                # Check btag
                if rstb_btag != b[1]:
                    raise UsbtmcException("Read status byte btag mismatch", 'read_stb')

                if self.interrupt_in_ep is None:
                    # No interrupt channel, value is here.
                    return b[2]
                else:
                    # Read response from interrupt channel.
                    resp = self.interrupt_in_ep.read(2, timeout=self._timeout_ms)
                    if resp[0] != rstb_btag + 128:
                        raise UsbtmcException("Read status byte btag mismatch", 'read_stb')
                    else:
                        return resp[1]
            else:
                raise UsbtmcException("Read status failed", 'read_stb')
        else:
            return int(self.ask("*STB?"))

    def trigger(self):
        """Send trigger command."""
        # Open device if needed.
        if not self.connected:
            self.open()

        # Check which protocol to use.
        if self.support_trigger:
            # USB488 protocol.
            data = self.pack_usb488_trigger()
            self.bulk_out_ep.write(data, timeout=self._timeout_ms)
        else:
            self.write("*TRG")

    def clear(self):
        """Send clear command."""
        # Open device if needed.
        if not self.connected:
            self.open()

        # Send INITIATE_CLEAR
        b = self.device.ctrl_transfer(
            bmRequestType=usb.util.build_request_type(usb.util.CTRL_IN, usb.util.CTRL_TYPE_CLASS, usb.util.CTRL_RECIPIENT_INTERFACE),
            bRequest=USBTMC_REQUEST_INITIATE_CLEAR,
            wValue=0x0000,
            wIndex=self.iface.index,
            data_or_wLength=0x0001,
            timeout=self._timeout_ms
        )

        # Check response.
        if (b[0] == USBTMC_STATUS_SUCCESS):
            # Initiate clear succeeded, wait for completion
            while True:
                # Request status.
                b = self.device.ctrl_transfer(
                    bmRequestType=usb.util.build_request_type(usb.util.CTRL_IN, usb.util.CTRL_TYPE_CLASS, usb.util.CTRL_RECIPIENT_INTERFACE),
                    bRequest=USBTMC_REQUEST_CHECK_CLEAR_STATUS,
                    wValue=0x0000,
                    wIndex=self.iface.index,
                    data_or_wLength=0x0002,
                    timeout=self._timeout_ms
                )

                # Check for completion.
                time.sleep(0.1)
                if (b[0] != USBTMC_STATUS_PENDING):
                    break

            # Clear halt condition
            self.bulk_out_ep.clear_halt()
            if self.force_clear_bulk_in:
                self.bulk_in_ep.clear_halt()
        else:
            raise UsbtmcException("Clear failed", 'clear')

    def _abort_bulk_out(self, btag=None):
        """Abort bulk out."""
        # Check if device is open (otherwise nothing to abort).
        if not self.connected:
            return

        # Check which bulk tag to use.
        if btag is None:
            btag = self.last_btag

        # Send INITIATE_ABORT_BULK_OUT
        abort_timeout_ms = int(1000 * self.abort_timeout)
        b = self.device.ctrl_transfer(
            bmRequestType=usb.util.build_request_type(usb.util.CTRL_IN, usb.util.CTRL_TYPE_CLASS, usb.util.CTRL_RECIPIENT_ENDPOINT),
            bRequest=USBTMC_REQUEST_INITIATE_ABORT_BULK_OUT,
            wValue=btag,
            wIndex=self.bulk_out_ep.bEndpointAddress,
            data_or_wLength=0x0002,
            timeout=abort_timeout_ms
        )

        # Parse response.
        if (b[0] == USBTMC_STATUS_SUCCESS):
            # Initiate abort bulk out succeeded, wait for completion
            while True:
                # Request status.
                b = self.device.ctrl_transfer(
                    bmRequestType=usb.util.build_request_type(usb.util.CTRL_IN, usb.util.CTRL_TYPE_CLASS, usb.util.CTRL_RECIPIENT_ENDPOINT),
                    bRequest=USBTMC_REQUEST_CHECK_ABORT_BULK_OUT_STATUS,
                    wValue=0x0000,
                    wIndex=self.bulk_out_ep.bEndpointAddress,
                    data_or_wLength=0x0008,
                    timeout=abort_timeout_ms
                )

                # Check for completion.
                time.sleep(0.1)
                if (b[0] != USBTMC_STATUS_PENDING):
                    break

        # Check result.
        if (b[0] == USBTMC_STATUS_SUCCESS):
            # Abort request completed. Clear endpoint.
            # See USBTMC v1.00 4.2.1.3:
            #     "The host must send a CLEAR_FEATURE control endpoint
            #     request to clear the Bulk-OUT halt."
            self.bulk_out_ep.clear_halt()

    def _abort_bulk_in(self, btag=None):
        """Abort bulk in."""
        # Check if device is open (otherwise nothing to abort).
        if not self.connected:
            return

        # Check which bulk tag to use.
        if btag is None:
            btag = self.last_btag

        # Send INITIATE_ABORT_BULK_IN
        abort_timeout_ms = int(1000 * self.abort_timeout)
        b = self.device.ctrl_transfer(
            bmRequestType=usb.util.build_request_type(usb.util.CTRL_IN, usb.util.CTRL_TYPE_CLASS, usb.util.CTRL_RECIPIENT_ENDPOINT),
            bRequest=USBTMC_REQUEST_INITIATE_ABORT_BULK_IN,
            wValue=btag,
            wIndex=self.bulk_in_ep.bEndpointAddress,
            data_or_wLength=0x0002,
            timeout=abort_timeout_ms
        )

        # Check result.
        if (b[0] == USBTMC_STATUS_SUCCESS):
            # Read remaining data from bulk in endpoint.
            # This is a required step before the abort request can complete.
            #
            # See USBTMC v1.00 4.2.1.4:
            #     "The host should continue reading from the Bulk-IN endpoint
            #     until a short packet is received."
            # USBTMC v1.00 4.2.1.5:
            #     "The host should not send CHECK_ABORT_BULK_IN_STATUS until
            #     a short Bulk-IN packet has been received."
            resp = self.bulk_in_ep.read(self.max_transfer_size, timeout=abort_timeout_ms)

            # Initiate abort bulk in succeeded, wait for completion
            while True:
                # Request status.
                b = self.device.ctrl_transfer(
                    bmRequestType=usb.util.build_request_type(usb.util.CTRL_IN, usb.util.CTRL_TYPE_CLASS, usb.util.CTRL_RECIPIENT_ENDPOINT),
                    bRequest=USBTMC_REQUEST_CHECK_ABORT_BULK_IN_STATUS,
                    wValue=0x0000,
                    wIndex=self.bulk_in_ep.bEndpointAddress,
                    data_or_wLength=0x0008,
                    timeout=abort_timeout_ms
                )

                # Check for completion.
                time.sleep(0.1)
                if (b[0] != USBTMC_STATUS_PENDING):
                    break
        else:
            # no transfer in progress; nothing to do
            pass

    def remote(self):
        """Send remote command."""
        raise NotImplementedError()

    def local(self):
        """Send local command."""
        raise NotImplementedError()

    def lock(self):
        """Send lock command (Avantest only)."""
        # Open device if needed.
        if not self.connected:
            self.open()

        if self.advantest_quirk:
            # This Advantest/ADCMT vendor-specific control command enables remote control and must be sent before any
            # commands are exchanged (otherwise READ commands will only retrieve the latest measurement).
            self.advantest_locked = True
            self.device.ctrl_transfer(
                bmRequestType=0xA1, bRequest=0xA0, wValue=0x0001, wIndex=0x0000, data_or_wLength=1
            )
        else:
            raise NotImplementedError()

    def unlock(self):
        """Send unlock command (Avantest only)."""
        # Open device if needed.
        if not self.connected:
            self.open()

        if self.advantest_quirk:
            # This Advantest/ADCMT vendor-specific control command enables remote control and must be sent before any
            # commands are exchanged (otherwise READ commands will only retrieve the latest measurement).
            self.advantest_locked = False
            self.device.ctrl_transfer(
                bmRequestType=0xA1, bRequest=0xA0, wValue=0x0000, wIndex=0x0000, data_or_wLength=1
            )
        else:
            raise NotImplementedError()

    def advantest_read_myid(self):
        """Read MyID value from Advantest and ADCMT devices."""
        # Open device if needed.
        if not self.connected:
            self.open()

        if self.advantest_quirk:
            # This Advantest/ADCMT vendor-specific control command reads the "MyID" identifier
            try:
                r = self.device.ctrl_transfer(
                    bmRequestType=0xC1, bRequest=0xF5, wValue=0x0000, wIndex=0x0000, data_or_wLength=1
                )
                return int(r[0])
            except:
                return None
        else:
            raise NotImplementedError()

    def _release_kernel_driver(self, interface_number):
        """Release kernel driver (POSIX)."""
        if os.name == 'posix':
            if self.device.is_kernel_driver_active(interface_number):
                self.reattach.append(interface_number)
                try:
                    self.device.detach_kernel_driver(interface_number)
                except usb.core.USBError as e:
                    sys.exit(
                        "Could not detach kernel driver from interface({0}): {1}".format(interface_number,
                                                                                         str(e)))



"""
Imagine Eyes, deformable mirror.

The qmi.instruments.imagine_eyes package provides support for:
- Mirao 52-e electromagnetic deformable mirror.
"""

from qmi.instruments.imagine_eyes.mirao52e import ImagineEyes_Mirao52e, MirrorStatus, read_mirao_file



"""Routines for structured configuration data.

This module provides functions for loading configuration data into
a tree structure of Python data classes.
This provides two advantages:

 1. The configuration file is validated and type-checked against
    the expected structure immediately when the file is loaded.
 2. The configuration data can be accessed through attributes of well-typed
    data classes instead of through dictionary lookups.

The structure of the configuration data is defined as a set of data classes.
Each structure is defined using the ``@configstruct`` decorator.

Only the following data types are allowed for fields of configuration structures:

 - ``int``, ``float``, ``str``, ``bool``
 - another configuration structure type;
 - ``List[T]`` where `T` is a supported field type
 - ``Dict[str, T]`` where `T` is a supported field type
 - ``Tuple[T1, T2, T3]`` or ``Tuple[T, ...]`` where `T`, `T1`, etc are supported field types
 - ``Optional[T]`` where `T` is a supported field type
 - ``Any``

A default value may be specified in the data class definition.
The default value, if present, will be used when the field is not present
in the configuration data. Fields without default value are mandatory.

Existing structure definitions may be extended (for example at application
level). This is done by subclassing the existing definition and adding
more fields.
"""

import collections
import dataclasses

from typing import Any, Dict, List, Tuple, TypeVar, Type, Union

from qmi.core.exceptions import QMI_ConfigurationException

_T = TypeVar("_T")
_StrDict = Dict[str, Any]


def configstruct(cls: Type[_T]) -> Type[_T]:
    """Class decorator to mark classes that hold QMI configuration data.

    This decorator is based on the standard `@dataclass` decorator.
    It calls ``@dataclass(init=False)`` to do most of the work,
    then adds a custom `__init__()` method.
    """

    # New __init__() function for the class.
    def initfn(self: Any, **kwargs: Any) -> None:

        # Walk the list of field definitions.
        for f in dataclasses.fields(cls):  # type: ignore
            if f.init and f.name in kwargs:
                # Value specified as argument to __init__().
                value = kwargs.pop(f.name)
                _parse_config_value(value, f.type, [])
                setattr(self, f.name, value)
            elif f.default is not dataclasses.MISSING:
                # No value specified; use default value of field.
                setattr(self, f.name, f.default)
            elif f.default_factory is not dataclasses.MISSING:
                # No value specified; use default factory of field.
                setattr(self, f.name, f.default_factory())
            elif f.init:
                # No value specified and no default.
                raise TypeError(f"{cls.__name__}.__init__() missing required argument {f.name!r}")

        # Check for unexpected keyword parameters.
        if kwargs:
            for arg_name in kwargs:
                raise TypeError(
                    f"{cls.__name__}.__init__() got an unexpected keyword argument {arg_name!r}"
                )

    # Invoke dataclass decorator.
    cls = dataclasses.dataclass(init=False)(cls)

    # Add marker.
    cls.__qmi_configstruct__ = True  # type: ignore

    # Add __init__ method.
    cls.__init__ = initfn  # type: ignore

    # Return modified class.
    return cls


def config_struct_to_dict(cfg: Any) -> Dict[str, Any]:
    """Convert configuration data from a dataclass instance to a dict.

    The returned dict will be suitable for serialization to JSON.

    Parameters:
        cfg: Configuration data (an instance of a dataclass).

    Returns:
        The configuration data converted to a dictionary.
    """

    if not dataclasses.is_dataclass(type(cfg)):
        raise TypeError("Configuration data must be a dataclass instance")

    ret = _inner_config_struct_to_dict(cfg)

    return ret


def _inner_config_struct_to_dict(cfg: Any) -> Any:
    """Inner recursive function that converts config_struct to dict"""
    if dataclasses.is_dataclass(type(cfg)):
        return _dictify_dataclass(cfg)
    elif isinstance(cfg, (list, tuple)):
        return _dictify_list_value(cfg)
    elif isinstance(cfg, dict):
        return _dictify_dict_value(cfg)
    elif (cfg is None) or isinstance(cfg, (int, float, bool, str)):
        return cfg
    else:
        raise TypeError(f"Unsupported value type: {cfg!r}")


def _dictify_list_value(cfg_list: Union[list, tuple]) -> list:
    """Copy a list value while converting its elements from structured to dict."""
    return [_inner_config_struct_to_dict(v) for v in cfg_list]


def _dictify_dict_value(cfg_dict: dict) -> dict:
    """Copy a dict value while converting its values from structured to dict."""
    ret: collections.OrderedDict = collections.OrderedDict()
    for (key, value) in cfg_dict.items():
        if not isinstance(key, str):
            raise TypeError(f"Unsupported non-string dictionary key: {key!r}")
        ret[key] = _inner_config_struct_to_dict(value)
    return ret


def _dictify_dataclass(cfg_dataclass: Any) -> dict:
    """Copy a dict value while converting its elements from dataclass to dict."""
    ret: collections.OrderedDict = collections.OrderedDict()
    for field in dataclasses.fields(cfg_dataclass):
        if hasattr(cfg_dataclass, field.name):
            value = getattr(cfg_dataclass, field.name)
            ret[field.name] = _inner_config_struct_to_dict(value)
    return ret


def _parse_config_value(val: Any, field_type: Any, path: List[str]) -> Any:
    """Convert JSON value to expected type."""

    # Recognize Optional[T]
    optional = False
    if repr(field_type).startswith("typing.Union[") or repr(field_type).startswith("typing.Optional["):
        for t in field_type.__args__:
            if t == type(None):  # This is intentional! Do not change to 't is None' as it will fail
                optional = True
            else:
                field_type = t

    if field_type == Any:
        # This field allows any type; pass the actual value without conversion.
        return val

    if (val is None) and (optional or (field_type is None) or (field_type == type(None))):
        # This field accepts None and actual value is None; pass it.
        return val

    if (field_type in (int, float, str, bool)) and isinstance(val, field_type):
        # This field accepts the actual scalar value.
        return val

    if (field_type is float) and isinstance(val, int):
        # Implicit conversion of integer value to floating point.
        return float(val)

    # Recognize "List" and "Dict" and map them to "list" and "dict".
    if field_type == List:
        field_type = list
    elif field_type == Tuple:
        field_type = tuple
    elif field_type == Dict:
        field_type = dict

    # Recognize untyped "list", "tuple" and "dict" values.
    if (field_type in (list, tuple, dict)) and isinstance(val, field_type):
        return val

    # Recognize untyped "tuple" values that are actually list instances.
    if (field_type is tuple) and isinstance(val, list):
        return tuple(val)

    # Recognize List[T].
    type_repr = repr(field_type)
    if type_repr.startswith("typing.List["):
        (elem_type,) = field_type.__args__
        if isinstance(val, list):
            ret = []
            for (i, elem) in enumerate(val):
                path.append("[{}]".format(i))
                ret.append(_parse_config_value(elem, elem_type, path))
                path.pop()
            return ret

    # Recognize Tuple[...].
    if type_repr.startswith("typing.Tuple["):
        if (len(field_type.__args__) == 2) and (field_type.__args__[1] is Ellipsis):
            elem_type = field_type.__args__[0]
            if isinstance(val, (list, tuple)):
                ret = []
                for (i, elem) in enumerate(val):
                    path.append("[{}]".format(i))
                    ret.append(_parse_config_value(elem, elem_type, path))
                    path.pop()
                return tuple(ret)
        elif len(val) == len(field_type.__args__):
            if isinstance(val, (list, tuple)) and (len(val) == len(field_type.__args__)):
                ret = []
                for (i, elem) in enumerate(val):
                    path.append("[{}]".format(i))
                    ret.append(_parse_config_value(elem, field_type.__args__[i], path))
                    path.pop()
                return tuple(ret)

    # Recognize Dict[str, T].
    if type_repr.startswith("typing.Dict[") and isinstance(val, dict):
        return _parse_config_dict(val, field_type, path)

    # Recognize data class.
    if dataclasses.is_dataclass(field_type):
        if isinstance(val, dict):
            return _parse_config_struct(val, field_type, path)
        elif dataclasses.is_dataclass(val) and not isinstance(val, type):
            return _parse_config_struct(dataclasses.asdict(val), field_type, path)

    pathstr = ".".join(path)
    raise QMI_ConfigurationException(
        f"Type mismatch in configuration item {pathstr}: got {type(val)} while expecting {field_type}"
    )


def _parse_config_dict(val: dict, field_type: Any, path: List[str]) -> _StrDict:
    """Parse Config Dict"""

    (_, elem_type) = field_type.__args__
    ret = collections.OrderedDict()  # type: _StrDict
    for (k, elem) in val.items():
        if not isinstance(k, str):
            pathstr = ".".join(path)
            raise QMI_ConfigurationException(
                f"Unsupported non-string dictionary key {k!r} in configuration item {pathstr}"
            )
        path.append("[{!r}]".format(k))
        ret[k] = _parse_config_value(elem, elem_type, path)
        path.pop()
    return ret


def _parse_config_struct(data: Any, cls: Any, path: List[str]) -> Any:
    """Convert dictionary to dataclass instance."""

    items = collections.OrderedDict()  # type: _StrDict

    # Walk the list of field definitions.
    for f in dataclasses.fields(cls):
        if f.init and (f.name in data):
            # Value specified for this field.
            path.append(f.name)
            items[f.name] = _parse_config_value(data[f.name], f.type, path)
            path.pop()
        elif f.init and (f.default is dataclasses.MISSING)\
                and (f.default_factory is dataclasses.MISSING):  # type: ignore
            # Value not specified and no default.
            path.append(f.name)
            pathstr = ".".join(path)
            raise QMI_ConfigurationException(f"Missing value for required configuration item {pathstr}")

    # Check for left-over fields.
    for k in data.keys():
        if k not in items:
            path.append(k)
            pathstr = ".".join(path)
            raise QMI_ConfigurationException(f"Unknown configuration item {pathstr}")

    # Construct dataclass instance.
    return cls(**items)


def _check_config_struct_type(cls: Any, path: List[str]) -> None:
    """Check that the specified type is acceptable for configuration data."""

    pathstr = ".".join(path)

    # Recognize Optional[T]
    if repr(cls).startswith("typing.Union[") or repr(cls).startswith("typing.Optional["):
        nsub = 0
        for t in cls.__args__:
            if t == type(None):  # This is intentional! Do not change to 't is None' as it will fail
                pass
            elif nsub == 0:
                cls = t
                nsub += 1
            else:
                raise QMI_ConfigurationException(f"Unsupported Union type in configuration field {pathstr}")

    # Recognize scalar types.
    if cls in (int, float, str, bool):
        return  # accept

    # Recognize wildcard.
    if cls == Any:
        return  # accept

    # Recognize untyped aggregates.
    if cls in (list, dict, Any, List, Tuple, Dict):
        return  # accept

    # Recognize List[T].
    type_repr = repr(cls)
    if type_repr.startswith("typing.List["):
        (elem_type,) = cls.__args__
        path.append("[]")
        _check_config_struct_type(elem_type, path)
        path.pop()
        return  # accept

    # Recognize Tuple[...].
    if type_repr.startswith("typing.Tuple["):
        if (len(cls.__args__) == 2) and (cls.__args__[1] is Ellipsis):
            elem_type = cls.__args__[0]
            path.append("[]")
            _check_config_struct_type(elem_type, path)
            path.pop()
            return  # accept
        else:
            for (i, elem_type) in enumerate(cls.__args__):
                path.append("[{}]".format(i))
                _check_config_struct_type(elem_type, path)
                path.pop()
            return  # accept

    # Recognize Dict[str, T].
    if type_repr.startswith("typing.Dict["):
        (key_type, elem_type) = cls.__args__
        if key_type is not str:
            raise QMI_ConfigurationException(
                f"Unsupported non-string-key dictionary type in configuration field {pathstr}"
            )
        path.append("[]")
        _check_config_struct_type(elem_type, path)
        path.pop()
        return  # accept

    # Recognize data class.
    if dataclasses.is_dataclass(cls) and isinstance(cls, type):
        # Check each field type.
        for f in dataclasses.fields(cls):
            if f.init:
                path.append(f.name)
                _check_config_struct_type(f.type, path)
                path.pop()
        return  # accept

    # Reject.
    raise QMI_ConfigurationException(f"Unsupported data type in configuration field {pathstr}")


def config_struct_from_dict(data: _StrDict, cls: Type[_T]) -> _T:
    """Convert configuration data from a dictionary to a dataclass instance.

    The input data is typically obtained by parsing a JSON file.

    Parameters:
        data: Configuration data as a dict instance.
        cls: Dataclass type to be used for holding the configuration data.

    Returns:
        An instance of the specified dataclass containing the configuration data.

    Raises:
        ~qmi.core.exceptions.QMI_ConfigurationException: If the
            configuration data does not match the expected structure.
    """

    if (not dataclasses.is_dataclass(cls)) or (not isinstance(cls, type)):
        raise TypeError("Configuration class type must be a dataclass")

    # Verify that the dataclass type contains only supported field types.
    _check_config_struct_type(cls, [])

    # Convert dictionary items to structure field assignments.
    return _parse_config_struct(data, cls, [])



import struct
import unittest, unittest.mock

from qmi.instruments.thorlabs import Thorlabs_Mff10X

import qmi.core.exceptions
from qmi.core.context import QMI_Context


class TestParsingAndFormatFunctions(unittest.TestCase):

    def test_format_msg_hw_req_info(self):
        # arrange
        dest = 0x1
        source = 0x2
        expected = bytes([0x05, 0x00, 0x00, 0x00, dest, source])
        # act
        msg = qmi.instruments.thorlabs.mff10x._format_msg_hw_req_info(dest, source)
        # assert
        self.assertEqual(msg, expected)

    def test_format_msg_mot_req_statusbits(self):
        # arrange
        chan_iden = 0x3
        dest = 0x1
        source = 0x2
        expected = bytes([0x29, 0x04, chan_iden, 0x00, dest, source])
        # act
        msg = qmi.instruments.thorlabs.mff10x._format_msg_mot_req_statusbits(chan_iden, dest, source)
        # assert
        self.assertEqual(msg, expected)

    def test_format_msg_mot_move_jog(self):
        # arrange
        chan_iden = 0x3
        direction = 0x4
        dest = 0x1
        source = 0x2
        expected = bytes([0x6A, 0x04, chan_iden, direction, dest, source])
        # act
        msg = qmi.instruments.thorlabs.mff10x._format_msg_mot_move_jog(chan_iden, direction, dest, source)
        # assert
        self.assertEqual(msg, expected)

    def test_parse_msg_hw_get_info(self):
        """Test that parsing works as expected to extract hardware info"""
        # arrange
        serial_number = "\x01\x02\x03\x04"  # bits 6:10
        model_number = "5678901\x00"  # bits 10:18
        hw_type = "\x01\x00"  # bits 18-19
        fw_version = "\x02\x03\x04"  # bits 22, 21, 20)
        hw_version, mod_state, num_channels = "\x05\x00", "\x06\x00", "\x07\x00"  # bits 84-85, 86-87, 88-89
        fill_in_middle = "\x03\x04\x05\x06\x07\x08\x09\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x00\x01\x02\x03\x04" + \
            "\x05\x06\x07\x08\x09\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x00\x01\x02\x03\x04\x05\x06\x07\x08" + \
            "\x09\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x00\x01\x02\x03"
        end_fill = "\x00"  # bit 90
        input_string = "\x06\x00\x5a\x00\x81" + serial_number + model_number + hw_type + fw_version +\
            fill_in_middle + hw_version + mod_state + num_channels + end_fill
        input_msg = bytearray(input_string.encode())

        expected_serial_number = struct.unpack('<I', bytearray(serial_number.encode()))[0]
        expected_model_number = model_number.rstrip("\x00")
        expected_hw_type = struct.unpack('<H', bytearray(hw_type.encode()))[0]
        expected_fw_version = (ord(str(fw_version[2])), ord(str(fw_version[1])), ord(str(fw_version[0])))
        expected_hw_version = struct.unpack('<H', bytearray(hw_version.encode()))[0]
        expected_mod_state = struct.unpack('<H', bytearray(mod_state.encode()))[0]
        expected_num_channels = struct.unpack('<H', bytearray(num_channels.encode()))[0]
        # act
        hwinfo = qmi.instruments.thorlabs.mff10x._parse_msg_hw_get_info(input_msg)
        # assert
        self.assertEqual(hwinfo.serial_number, expected_serial_number)
        self.assertEqual(hwinfo.model_number, expected_model_number)
        self.assertEqual(hwinfo.hw_type, expected_hw_type)
        self.assertTupleEqual(hwinfo.fw_version, expected_fw_version)
        self.assertEqual(hwinfo.hw_version, expected_hw_version)
        self.assertEqual(hwinfo.mod_state, expected_mod_state)
        self.assertEqual(hwinfo.num_channels, expected_num_channels)

    def test_parse_msg_hw_get_info_raises_exceptions(self):
        """Test that parsing message checks works as expected to raise exceptions"""
        # arrange
        input_string_1 = "\x64\x04\x5a\x00\x81\x00"
        expected_message_1 = "Got message ID {:04x} while expecting MSG_HW_GET_INFO".format(
            struct.unpack('<H', bytearray(input_string_1[:2].encode()))[0]
            )
        input_string_2 = "\x06\x00\x5a\x00\x81\x00"
        expected_message_2 = "Got {} bytes while expecting 90 bytes MSG_HW_GET_INFO".format(len(input_string_2) + 1)
        # act
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc_1:
            qmi.instruments.thorlabs.mff10x._parse_msg_hw_get_info(bytearray(input_string_1.encode()))

        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc_2:
            qmi.instruments.thorlabs.mff10x._parse_msg_hw_get_info(bytearray(input_string_2.encode()))

        # assert
        self.assertEqual(exc_1.exception.args[0], expected_message_1)
        self.assertEqual(exc_2.exception.args[0], expected_message_2)

    def test_parse_msg_mot_get_statusbits(self):
        """Test getting status bits works"""
        # arrange
        status_bits = "\x00\x00\x01\x02\x03\x04"
        input_string = "\x2A\x04\x5a\x00\x81" + status_bits
        expected = struct.unpack('<I', bytearray(status_bits[-4:].encode()))[0]
        # act
        res = qmi.instruments.thorlabs.mff10x._parse_msg_mot_get_statusbits(bytearray(input_string.encode()))
        # assert
        self.assertEqual(res, expected)

    def test_parse_msg_mot_get_statusbits_raises_exceptions(self):
        """Test getting status bits raises exceptions"""
        # arrange
        input_string_1 = "\x04\x2A\x5a\x00\x81"
        expected_message_1 = "Got message ID {:04x} while expecting MSG_MOT_GET_STATUSBITS".format(
            struct.unpack('<H', bytearray(input_string_1[:2].encode()))[0]
            )
        input_string_2 = "\x2A\x04\x5a\x00\x81"
        expected_message_2 = "Got {} bytes while expecting 12 bytes MSG_MOT_GET_STATUSBITS".format(
            len(input_string_2) + 1)
        # act
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc_1:
            qmi.instruments.thorlabs.mff10x._parse_msg_mot_get_statusbits(bytearray(input_string_1.encode()))

        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc_2:
            qmi.instruments.thorlabs.mff10x._parse_msg_mot_get_statusbits(bytearray(input_string_2.encode()))

        # assert
        self.assertEqual(exc_1.exception.args[0], expected_message_1)
        self.assertEqual(exc_2.exception.args[0], expected_message_2)

    def test_parse_msg_mot_move_completed(self):
        """Test that correct move response raises no exception"""
        input_string = "\x64\x04\x5a\x00\x81\x00"
        qmi.instruments.thorlabs.mff10x._parse_msg_mot_move_completed(bytearray(input_string.encode()))

    def test_parse_msg_mot_move_completed_raises_exception(self):
        """Test that wrong move response raises an exception"""
        input_string = "\x04\x64\x5a\x00\x81\x00"
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
            qmi.instruments.thorlabs.mff10x._parse_msg_mot_move_completed(bytearray(input_string.encode()))


class TestThorlabsMFF10x(unittest.TestCase):

    def setUp(self) -> None:
        unittest.mock.patch("qmi.core.transport.QMI_SerialTransport._validate_device_name")
        qmi_context = unittest.mock.MagicMock(spec=QMI_Context)
        qmi_context.name = "mockyflop"
        self.ser_address = "COM100"
        self.baudrate = 115200
        transport_id = "serial:{}:baudrate={}".format(self.ser_address, self.baudrate)
        self.thorlabs = Thorlabs_Mff10X(qmi_context, "flippy", transport_id)

    def test_open_close(self):
        with unittest.mock.patch("serial.Serial") as ser:
            self.thorlabs.open()
            self.thorlabs.close()

            ser.assert_called_once_with(
                self.ser_address,
                baudrate=self.baudrate,  # The rest are defaults
                bytesize=8,
                parity='N',
                rtscts=True,
                stopbits=1.0,
                timeout=0.04
                )

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read")
    def test_read_message(self, mock_read):
        """See that the _read_message method reads only the expected 6 bits. Fill the buffer with `random` bits"""
        mock_read.return_value = bytearray("blablabla".encode())  # returns 9 bits
        expected = bytearray("blabla".encode())  # but the regular return with extra data is of expected length of 6.
        self.thorlabs._transport.read.side_effect = [expected[:6], expected[6:22]]
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            msg = self.thorlabs._read_message(None)
            self.thorlabs.close()

        self.assertEqual(msg, expected)

    def test_read_message_with_data(self):
        """6-bit input `00x10x00x81a` should trigger extra data read of 16 bits in _read_buffer call."""
        # 5th bit 0x81 indicates that extra data is present in buffer, and 3rd bit 0x10 that it is 16 bits of length.
        expected = bytearray("00\x10\x00\x81abfdaj;jb;aj;lf;jl".encode())  # Total length of 24 bits
        self.thorlabs._transport.read = unittest.mock.MagicMock(self.thorlabs._transport.read)
        self.thorlabs._transport.read.side_effect = [expected[:6], expected[6:22]]
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            msg = self.thorlabs._read_message(None)
            self.thorlabs.close()

        self.assertEqual(msg, expected[:(6+16)])  # We should see total of 22 bits read

    def test_read_message_with_data_invalid_length(self):
        """"An error should be raised if the extra data present is not of expected length"""
        expected = bytearray("blab\x81a123456.7".encode())  # bit 3 'a' indicates 97 extra data bits, but we have only 8
        self.thorlabs._transport.read = unittest.mock.MagicMock(self.thorlabs._transport.read)
        self.thorlabs._transport.read.side_effect = [expected[:6], expected[6:]]
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.thorlabs._read_message(None)

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read")
    def test_get_idn(self, mock_read):
        # arrange
        serial_number = "\x01\x02\x03\x04"  # bits 6:10
        model_number = "5678901\x00"  # bits 10:18
        hw_type = "\x01\x00"  # bits 18-19
        fw_version = "\x02\x03\x04"  # bits 22, 21, 20)
        hw_version, mod_state, num_channels = "\x05\x00", "\x06\x00", "\x07\x00"  # bits 84-85, 86-87, 88-89
        fill_in_middle = "\x03\x04\x05\x06\x07\x08\x09\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x00\x01\x02\x03\x04" + \
                         "\x05\x06\x07\x08\x09\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x00\x01\x02\x03\x04\x05\x06" + \
                         "\x07\x08\x09\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x00\x01\x02\x03"
        end_fill = "\x00"  # bit 90
        input_string = "\x06\x00\x5a\x00\x81" + serial_number + model_number + hw_type + fw_version + \
                       fill_in_middle + hw_version + mod_state + num_channels + end_fill

        expected = 'QMI_InstrumentIdentification(vendor=\'Thorlabs\', model=\'{}\', serial=\'{}\', version=\'{}.{}.{}\')'.format(
            model_number.rstrip("\x00"), struct.unpack('<I', bytearray(serial_number.encode()))[0],
                ord(str(fw_version[2])), ord(str(fw_version[1])), ord(str(fw_version[0])))
        mock_read.return_value = bytearray(input_string.encode())
        # act
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            idn = self.thorlabs.get_idn()
            self.thorlabs.close()

        # assert
        self.assertEqual(str(idn), expected)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read")
    def test_get_position(self, mock_read):
        """Test that get_position() returns a set of random status bits for position."""
        # arrange
        status_bits = "\x00\x00\x01\x02\x03\x04"  # Some random bits to represent the "status" reply
        input_string = "\x2A\x04\x5a\x00\x81" + status_bits  # 0x2A + 0x04 = 0x042A is the expected return code,
        # 0x81 indicates extra data present. The first two following are skipped (apparently)
        expected = struct.unpack('<I', bytearray(status_bits[-4:].encode()))[0] & 3
        mock_read.return_value = bytearray(input_string.encode())
        # act
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            pos = self.thorlabs.get_position()
            self.thorlabs.close()

        # assert
        self.assertEqual(pos, expected)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read")
    def test_move_mount_wait_complete(self, mock_read):
        """Test that move_mount() returns. A set of random status bits needed for the get_position() call within."""
        # arrange
        status_bits = "\x00\x00\x01\x02\x03\x04"  # Some random bits to represent the "status" reply, except for 0x01
        # that tells we are going into direction 1
        input_string = "\x2A\x04\x5a\x00\x81" + status_bits  # 0x2A + 0x04 = 0x042A is the expected return code
        mock_read.return_value = bytearray(input_string.encode())
        # act
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            self.thorlabs.move_mount(1, True)  # using max timeout
            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read")
    def test_move_mount_wait_complete_timeout(self, mock_read):
        """Test that get_position() time-outs, when no extra data is returned within time-out."""
        # arrange
        input_string = "\x2A\x04\x5a\x00\x81"  # 0x2A + 0x04 = 0x042A is the expected return code
        mock_read.return_value = bytearray(input_string.encode())
        # act and assert
        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException), unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            self.thorlabs.move_mount(1, True, 0.1)

        self.thorlabs.close()

    def test_move_mount_invalid_direction(self):
        """Test that ValueError is raised when calling 'move_mount' with invalid direction number."""
        # arrange, act and assert
        with self.assertRaises(ValueError), unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            self.thorlabs.move_mount(3, True)

        self.thorlabs.close()


if __name__ == '__main__':
    unittest.main()
